import contextlib
from typing import Optional, Type, Union

import pytest
import torch
from marshmallow import ValidationError

from ludwig.features.category_feature import CategoryOutputFeature
from ludwig.features.set_feature import SetOutputFeature
from ludwig.features.text_feature import TextOutputFeature
from ludwig.modules import loss_modules
from ludwig.schema.features.loss.loss import (
    BWCEWLossConfig,
    CORNLossConfig,
    HuberLossConfig,
    MAELossConfig,
    MAPELossConfig,
    MSELossConfig,
    RMSELossConfig,
    RMSPELossConfig,
    SigmoidCrossEntropyLossConfig,
    SoftmaxCrossEntropyLossConfig,
)
from ludwig.schema.model_config import ModelConfig
from tests.integration_tests.utils import category_feature, set_feature, text_feature


def from_float(v: float) -> torch.Tensor:
    return torch.tensor(v).float()


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(36).float()])
def test_mse_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.MSELoss(MSELossConfig())
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(6).float()])
def test_mae_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.MAELoss(MAELossConfig())
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(0.7365440726280212)])
def test_mape_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.MAPELoss(MAPELossConfig())
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(6).float()])
def test_rmse_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.RMSELoss(RMSELossConfig())
    assert loss(preds, target) == output


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(0.7527).float()])
def test_rmspe_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.RMSPELoss(RMSPELossConfig())
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.tensor([[0.1, 0.2]]).float()])
@pytest.mark.parametrize("target", [torch.tensor([[0.0, 0.2]]).float()])
@pytest.mark.parametrize("output", [torch.tensor(707.1068).float()])
def test_rmspe_loss_zero_targets(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.RMSPELoss(RMSPELossConfig())
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


@pytest.mark.parametrize(
    "confidence_penalty,positive_class_weight,robust_lambda,output",
    [
        (0.0, None, 0, from_float(-21.4655)),
        (2.0, None, 0, from_float(-21.1263)),
        (0.0, 2.0, 0, from_float(-20.1222)),
        (0.0, None, 2, from_float(22.4655)),
        (2, 2, 2, from_float(21.4614)),
    ],
)
@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
def test_bwcew_loss(
    preds: torch.Tensor,
    target: torch.Tensor,
    confidence_penalty: float,
    positive_class_weight: Optional[float],
    robust_lambda: int,
    output: torch.Tensor,
):
    loss = loss_modules.BWCEWLoss(
        BWCEWLossConfig(
            positive_class_weight=positive_class_weight,
            robust_lambda=robust_lambda,
            confidence_penalty=confidence_penalty,
        )
    )
    assert torch.isclose(loss(preds, target), output)


@pytest.mark.parametrize("preds", [torch.tensor([[0.5, 0.5], [0.2, 0.8], [0.6, 0.4]])])
@pytest.mark.parametrize("target", [torch.tensor([1, 1, 0])])
@pytest.mark.parametrize("output", [torch.tensor(0.5763)])
def test_softmax_cross_entropy_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.SoftmaxCrossEntropyLoss(SoftmaxCrossEntropyLossConfig())
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
@pytest.mark.parametrize("output", [torch.tensor(-21.4655).float()])
def test_sigmoid_cross_entropy_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.SigmoidCrossEntropyLoss(SigmoidCrossEntropyLossConfig())
    assert torch.isclose(loss(preds, target), output)


@pytest.mark.parametrize(
    "delta,output",
    [
        (1.0, from_float(5.5000)),
        (0.5, from_float(2.8750)),
        (2.0, from_float(10.0)),
        (0.0, ValidationError),
    ],
)
@pytest.mark.parametrize("preds", [torch.arange(6).reshape(3, 2).float()])
@pytest.mark.parametrize("target", [torch.arange(6, 12).reshape(3, 2).float()])
def test_huber_loss(
    preds: torch.Tensor, target: torch.Tensor, delta: float, output: Union[torch.Tensor, Type[Exception]]
):
    with pytest.raises(output) if not isinstance(output, torch.Tensor) else contextlib.nullcontext():
        loss = loss_modules.HuberLoss(HuberLossConfig.from_dict({"delta": delta}))
        value = loss(preds, target)
        assert value == output


@pytest.mark.parametrize("preds", [torch.tensor([[0.25, 0.2, 0.55], [0.2, 0.35, 0.45], [0.8, 0.1, 0.1]])])
@pytest.mark.parametrize("target", [torch.tensor([2, 1, 0])])
@pytest.mark.parametrize("output", [torch.tensor(0.7653)])
def test_corn_loss(preds: torch.Tensor, target: torch.Tensor, output: torch.Tensor):
    loss = loss_modules.CORNLoss(CORNLossConfig())
    assert torch.isclose(loss(preds, target), output, rtol=0.0001)


def test_dict_class_weights_category():
    input_features = [text_feature()]
    output_features = [category_feature(decoder={"vocab_size": 3})]
    config = {
        "input_features": input_features,
        "output_features": output_features,
    }

    # Set class weights as dictionary on config
    class_weights_dict = {"token_1": 0.1, "token_2": 0.2, "token_3": 0.3}
    config["output_features"][0]["loss"] = {"type": "softmax_cross_entropy", "class_weights": class_weights_dict}

    # Mock feature metadata
    feature_metadata = {
        "idx2str": ["token_1", "token_2", "token_3"],
        "str2idx": {"token_1": 0, "token_2": 1, "token_3": 2},
        "str2freq": {"token_1": 300, "token_2": 200, "token_3": 100},
        "vocab_size": 3,
        "preprocessing": {
            "missing_value_strategy": "drop_row",
            "fill_value": "<UNK>",
            "computed_fill_value": "<UNK>",
            "lowercase": False,
            "most_common": 10000,
            "cache_encoder_embeddings": False,
        },
    }

    model_config = ModelConfig.from_dict(config)

    CategoryOutputFeature.update_config_with_metadata(
        feature_config=model_config.output_features[0],
        feature_metadata=feature_metadata,
    )

    assert model_config.output_features[0].loss.class_weights == [0.1, 0.2, 0.3]


def test_dict_class_weights_text():
    input_features = [text_feature()]
    output_features = [text_feature(decoder={"vocab_size": 3, "max_sequence_length": 10})]
    config = {
        "input_features": input_features,
        "output_features": output_features,
    }

    # Set class weights as dictionary on config
    class_weights_dict = {
        "<EOS>": 0,
        "<SOS>": 0,
        "<PAD>": 0,
        "<UNK>": 0,
        "token_1": 0.5,
        "token_2": 0.4,
        "token_3": 0.1,
    }
    config["output_features"][0]["loss"] = {
        "type": "sequence_softmax_cross_entropy",
        "class_weights": class_weights_dict,
    }

    # Mock feature metadata
    feature_metadata = {
        "idx2str": ["<EOS>", "<SOS>", "<PAD>", "<UNK>", "token_1", "token_2", "token_3"],
        "str2idx": {"<EOS>": 0, "<SOS>": 1, "<PAD>": 2, "<UNK>": 3, "token_1": 4, "token_2": 5, "token_3": 6},
        "str2freq": {"<EOS>": 0, "<SOS>": 0, "<PAD>": 0, "<UNK>": 0, "token_1": 300, "token_2": 200, "token_3": 100},
        "str2idf": None,
        "vocab_size": 7,
        "max_sequence_length": 9,
        "max_sequence_length_99ptile": 9.0,
        "pad_idx": 2,
        "padding_symbol": "<PAD>",
        "unknown_symbol": "<UNK>",
        "index_name": None,
        "preprocessing": {
            "prompt": {
                "retrieval": {"type": None, "index_name": None, "model_name": None, "k": 0},
                "task": None,
                "template": None,
            },
            "pretrained_model_name_or_path": None,
            "tokenizer": "space_punct",
            "vocab_file": None,
            "sequence_length": None,
            "max_sequence_length": 256,
            "most_common": 20000,
            "padding_symbol": "<PAD>",
            "unknown_symbol": "<UNK>",
            "padding": "right",
            "lowercase": True,
            "missing_value_strategy": "drop_row",
            "fill_value": "<UNK>",
            "computed_fill_value": "<UNK>",
            "ngram_size": 2,
            "cache_encoder_embeddings": False,
            "compute_idf": False,
        },
    }

    model_config = ModelConfig.from_dict(config)

    TextOutputFeature.update_config_with_metadata(
        feature_config=model_config.output_features[0],
        feature_metadata=feature_metadata,
    )

    assert model_config.output_features[0].loss.class_weights == [0, 0, 0, 0, 0.5, 0.4, 0.1]


def test_dict_class_weights_set():
    input_features = [category_feature()]
    output_features = [set_feature()]
    config = {
        "input_features": input_features,
        "output_features": output_features,
    }

    # Set class weights as dictionary on config
    class_weights_dict = {"token_1": 0.1, "token_2": 0.2, "token_3": 0.3, "<UNK>": 0}
    config["output_features"][0]["loss"] = {"type": "sigmoid_cross_entropy", "class_weights": class_weights_dict}

    # Mock feature metadata
    feature_metadata = {
        "idx2str": ["token_1", "token_2", "token_3", "<UNK>"],
        "str2idx": {"token_1": 0, "token_2": 1, "token_3": 2, "<UNK>": 3},
        "str2freq": {"token_1": 300, "token_2": 200, "token_3": 100, "<UNK>": 0},
        "vocab_size": 4,
        "max_set_size": 3,
        "preprocessing": {
            "tokenizer": "space",
            "missing_value_strategy": "drop_row",
            "fill_value": "<UNK>",
            "computed_fill_value": "<UNK>",
            "lowercase": False,
            "most_common": 10000,
        },
    }

    model_config = ModelConfig.from_dict(config)

    SetOutputFeature.update_config_with_metadata(
        feature_config=model_config.output_features[0],
        feature_metadata=feature_metadata,
    )

    assert model_config.output_features[0].loss.class_weights == [0.1, 0.2, 0.3, 0]



from dataclasses import field
from typing import Dict

from marshmallow import fields, ValidationError

import ludwig.schema.utils as schema_utils
from ludwig.api_annotations import DeveloperAPI
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class ProfilerConfig(schema_utils.BaseMarshmallowConfig):
    """Dataclass that holds profiling parameters for torch profile scheduler.

    The profiler will skip the first skip_first steps, then wait for wait steps, then do the warmup for the next warmup
    steps, then do the active recording for the next active steps and then repeat the cycle starting with wait steps.
    The optional number of cycles is specified with the repeat parameter, the zero value means that the cycles will
    continue until the profiling is finished.
    """

    wait: int = schema_utils.IntegerRange(
        default=1,
        min=0,
        description="The number of steps to wait profiling.",
    )

    warmup: int = schema_utils.IntegerRange(
        default=1,
        min=0,
        description="The number of steps for profiler warmup after waiting finishes.",
    )

    active: int = schema_utils.IntegerRange(
        default=3,
        min=0,
        description="The number of steps that are actively recorded. Values more than 10 wil dramatically slow down "
        "tensorboard loading.",
    )

    repeat: int = schema_utils.IntegerRange(
        default=5,
        min=0,
        description="The optional number of profiling cycles. Use 0 to profile the entire training run.",
    )

    skip_first: int = schema_utils.IntegerRange(
        default=0,
        min=0,
        max=100,
        description="The number of steps to skip in the beginning of training.",
    )


@DeveloperAPI
def ProfilerDataclassField(description: str, default: Dict = {}):
    """Returns custom dataclass field for `ludwig.modules.profiler.ProfilerConfig`. Allows `None` by default.

    :param description: Description of the torch profiler field
    :param default: dict that specifies clipping param values that will be loaded by its schema class (default: {}).
    """
    allow_none = True

    class ProfilingMarshmallowField(fields.Field):
        """Custom marshmallow field class for the torch profiler.

        Deserializes a dict to a valid instance of `ludwig.modules.optimization_modules.ProfilerConfig` and
        creates a corresponding JSON schema for external usage.
        """

        def _deserialize(self, value, attr, data, **kwargs):
            if value is None:
                return value
            if isinstance(value, dict):
                try:
                    return ProfilerConfig.Schema().load(value)
                except (TypeError, ValidationError):
                    raise ValidationError(f"Invalid params for profiling config: {value}, see ProfilerConfig class.")
            raise ValidationError("Field should be None or dict")

        def _jsonschema_type_mapping(self):
            return {
                **schema_utils.unload_jsonschema_from_marshmallow_class(ProfilerConfig),
                "title": "profiler_options",
                "description": description,
            }

    if not isinstance(default, dict):
        raise ValidationError(f"Invalid default: `{default}`")

    def load_default():
        return ProfilerConfig.Schema().load(default)

    dump_default = ProfilerConfig.Schema().dump(default)

    return field(
        metadata={
            "marshmallow_field": ProfilingMarshmallowField(
                allow_none=allow_none,
                load_default=load_default,
                dump_default=dump_default,
                metadata={
                    "description": description,
                    "parameter_metadata": None,
                },
            )
        },
        default_factory=load_default,
    )



from abc import ABC, abstractmethod

import numpy as np


class VectorIndex(ABC):
    @abstractmethod
    def search(self, query: np.ndarray, k: int) -> np.ndarray:
        pass

    @abstractmethod
    def save(self, path: str):
        pass

    @classmethod
    @abstractmethod
    def from_path(cls, path: str) -> "VectorIndex":
        pass

    @classmethod
    @abstractmethod
    def from_embeddings(cls, embeddings: np.ndarray) -> "VectorIndex":
        pass



import os

import pytest

from ludwig.constants import COMBINER, EPOCHS, INPUT_FEATURES, OUTPUT_FEATURES, TRAINER, TYPE
from tests.integration_tests.utils import binary_feature, generate_data, run_test_suite, text_feature


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_text_adapter_lora(tmpdir, backend, ray_cluster_2cpu):
    input_features = [
        text_feature(
            encoder={
                "type": "auto_transformer",
                "pretrained_model_name_or_path": "hf-internal-testing/tiny-bert-for-token-classification",
                "trainable": True,
                "adapter": {"type": "lora"},
            },
        ),
    ]
    output_features = [binary_feature()]

    data_csv_path = os.path.join(tmpdir, "dataset.csv")
    dataset = generate_data(input_features, output_features, data_csv_path)

    config = {
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        COMBINER: {TYPE: "concat", "output_size": 14},
        TRAINER: {EPOCHS: 1},
    }
    model = run_test_suite(config, dataset, backend)

    state_dict = model.model.state_dict()

    # check that at least one of the keys contains the word "lora_" denoting a lora parameter
    assert any("lora_" in key for key in state_dict.keys())



from typing import Any, Dict, Optional

import pytest

from ludwig.constants import MODEL_ECD, TEXT, TYPE
from ludwig.schema.encoders.text_encoders import BERTConfig
from ludwig.schema.encoders.utils import get_encoder_cls
from ludwig.schema.features.preprocessing.text import TextPreprocessingConfig


@pytest.mark.parametrize(
    "pretrained_model_name_or_path",
    [None, "bert-large-uncased"],
    ids=["default_model", "override_model"],
)
def test_set_fixed_preprocessing_params(pretrained_model_name_or_path: str):
    expected_model_name = "bert-base-uncased"

    preprocessing = TextPreprocessingConfig.from_dict(
        {
            "tokenizer": "space",
            "lowercase": True,
        }
    )

    encoder_params = {}
    if pretrained_model_name_or_path is not None:
        encoder_params["pretrained_model_name_or_path"] = pretrained_model_name_or_path
        expected_model_name = pretrained_model_name_or_path

    encoder = BERTConfig.from_dict(encoder_params)
    encoder.set_fixed_preprocessing_params(MODEL_ECD, preprocessing)

    assert preprocessing.tokenizer == "hf_tokenizer"
    assert preprocessing.lowercase
    assert preprocessing.pretrained_model_name_or_path == expected_model_name


@pytest.mark.parametrize(
    "encoder_params,expected",
    [
        ({"type": "parallel_cnn"}, False),
        ({"type": "bert", "trainable": False}, True),
        ({"type": "bert", "trainable": True}, False),
    ],
    ids=["parallel_cnn", "bert_fixed", "bert_trainable"],
)
def test_set_fixed_preprocessing_params_cache_embeddings(encoder_params: Dict[str, Any], expected: Optional[bool]):
    preprocessing = TextPreprocessingConfig.from_dict(
        {
            "tokenizer": "space",
            "lowercase": True,
            "cache_encoder_embeddings": True,
        }
    )

    encoder = get_encoder_cls(MODEL_ECD, TEXT, encoder_params[TYPE]).from_dict(encoder_params)
    encoder.set_fixed_preprocessing_params(MODEL_ECD, preprocessing)
    assert preprocessing.cache_encoder_embeddings == expected



import logging
import re
from typing import Any, Dict, List, Union

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import CATEGORY, LOGITS, PREDICTIONS, PROBABILITIES, TEXT
from ludwig.decoders.base import Decoder
from ludwig.decoders.registry import register_decoder
from ludwig.decoders.utils import extract_generated_tokens
from ludwig.schema.decoders.llm_decoders import CategoryExtractorDecoderConfig, TextExtractorDecoderConfig
from ludwig.utils.strings_utils import get_tokenizer

logger = logging.getLogger(__name__)


# TODO(Arnav): Refactor to split into strategies like splitters
class Matcher:
    def __init__(self, match: Dict[str, Dict[str, Any]]):
        self.match = match

    def contains(self, decoded_input: str, value: str) -> bool:
        return value in decoded_input

    def regex(self, decoded_input: str, regex_pattern: str) -> bool:
        """Perform a regex match on a given text using a specified regex pattern.

        Parameters:
        text (str): The text to perform the match on.
        regex_pattern (str): The regex pattern to use for the match.

        Returns:
        A list of match objects.
        """
        # Compile the regex pattern
        matches = []
        try:
            regex = re.compile(regex_pattern)
            # Perform the match
            matches = regex.findall(decoded_input)
        except Exception:
            logger.warning(f"Regex pattern {regex_pattern} could not be compiled.")
        # If there is a match, matches is a non-empty list, so we can use this
        # to infer if there was a match or not and return a bool
        return len(matches) > 0

    def __call__(self, decoded_input: str) -> Union[str, None]:
        # Greedy match on first label that matches the input
        for label, label_def in self.match.items():
            label_def_type = label_def["type"]
            label_def_value = label_def["value"]

            if label_def_type == "contains":
                is_match = self.contains(decoded_input, label_def_value)
            elif label_def_type == "regex":
                is_match = self.regex(decoded_input, label_def_value)
            else:
                raise ValueError(
                    f"{label_def_type} is not a valid match `type`. Ludwig "
                    "currently supports `contains` and `regex` match types."
                )

            if is_match:
                return label
        return None


@DeveloperAPI
@register_decoder("text_extractor", [TEXT])
class TextExtractorDecoder(Decoder):
    def __init__(
        self,
        input_size: int,
        decoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = decoder_config
        self.input_size = input_size

        # Tokenizer
        self.tokenizer_type = self.config.tokenizer
        self.pretrained_model_name_or_path = self.config.pretrained_model_name_or_path
        self.vocab_file = self.config.vocab_file

        # Load tokenizer required for decoding the output from the generate
        # function of the text input feature for LLMs.
        self.tokenizer = get_tokenizer(self.tokenizer_type, self.vocab_file, self.pretrained_model_name_or_path)
        if hasattr(self.tokenizer, "tokenizer"):
            # Transformer Tokenizers
            self.tokenizer_vocab_size = self.tokenizer.tokenizer.vocab_size
        else:
            # TorchText Tokenizers
            self.tokenizer_vocab_size = len(self.tokenizer.vocab)

        # Maximum number of new tokens that will be generated
        # TODO(geoffrey): figure out where self.max_sequence_length is used– if not used, we might consider removing it.
        # It's confusing to have both this and `max_new_tokens` as a mandatory param in the `forward` function.
        self.max_sequence_length = self.config.max_new_tokens

    @staticmethod
    def get_schema_cls():
        return TextExtractorDecoderConfig

    @property
    def input_shape(self):
        return self.input_size

    def get_prediction_set(self):
        return {LOGITS, PREDICTIONS, PROBABILITIES}

    def forward(self, inputs: List[torch.Tensor], input_lengths: List[int], max_new_tokens: int):
        # Extract the sequences tensor from the LLMs forward pass
        generated_outputs = extract_generated_tokens(
            raw_generated_output_sequences=inputs,
            input_lengths=input_lengths,
            max_new_tokens=max_new_tokens,
            pad_sequence=True,
        )
        # Stack the predictions for each example in the batch. The padding should ensure they are all the same shape.
        for output in generated_outputs:
            if output.shape[0] > max_new_tokens:
                raise ValueError(
                    f"Output {output} is longer than the max_new_tokens {max_new_tokens} during decoding. "
                    f"This should never happen– please file an issue on GitHub."
                )

        generated_outputs = torch.stack(generated_outputs, dim=0)
        outputs_device = generated_outputs.device

        return {
            PREDICTIONS: generated_outputs,
            # TODO(Arnav): Add support for probabilities and logits
            PROBABILITIES: torch.zeros((len(generated_outputs), max_new_tokens, self.tokenizer_vocab_size)).to(
                outputs_device
            ),
            LOGITS: torch.zeros((len(generated_outputs), max_new_tokens, self.tokenizer_vocab_size)).to(outputs_device),
        }


@DeveloperAPI
@register_decoder("category_extractor", [CATEGORY])
class CategoryExtractorDecoder(Decoder):
    def __init__(
        self,
        decoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = decoder_config

        self.input_size = self.config.input_size
        self.fallback_label = self.config.fallback_label
        self.str2idx = self.config.str2idx
        self.vocab_size = len(self.config.str2idx)

        # Create Matcher object to perform matching on the decoded output
        self.matcher = Matcher(self.config.match)

        # Tokenizer
        self.tokenizer_type = self.config.tokenizer
        self.pretrained_model_name_or_path = self.config.pretrained_model_name_or_path
        self.vocab_file = self.config.vocab_file

        # Load tokenizer required for decoding the output from the generate
        # function of the text input feature for LLMs.
        self.tokenizer = get_tokenizer(self.tokenizer_type, self.vocab_file, self.pretrained_model_name_or_path)

    @staticmethod
    def get_schema_cls():
        return CategoryExtractorDecoderConfig

    @property
    def input_shape(self):
        return self.input_size

    def get_prediction_set(self):
        return {LOGITS, PREDICTIONS, PROBABILITIES}

    def forward(self, inputs: List[torch.Tensor], input_lengths: List[int], max_new_tokens: int):
        # Extract the sequences tensor from the LLMs forward pass
        generated_outputs = extract_generated_tokens(
            raw_generated_output_sequences=inputs,
            input_lengths=input_lengths,
            max_new_tokens=max_new_tokens,
            pad_sequence=False,
        )
        outputs_device = generated_outputs[0].device

        # Decode generated outputs from the LLM's generate function.
        decoded_outputs = self.tokenizer.tokenizer.batch_decode(generated_outputs, skip_special_tokens=True)

        # Parse labels based on matching criteria and return probability vectors
        matched_labels = []
        probabilities = []
        logits = []
        for output in decoded_outputs:
            output = output.lower()  # Convert to lowercase for matching

            matched_label = self.matcher(output)
            idx = self.str2idx[matched_label] if matched_label in self.str2idx else self.str2idx[self.fallback_label]

            # Append the index of the matched label
            matched_labels.append(idx)

            # Append the probability vector for the matched label
            probability_vec = [0] * self.vocab_size
            probability_vec[idx] = 1
            probabilities.append(probability_vec)

            # TODO(Arnav): Figure out how to compute logits. For now, we return
            # a tensor of zeros.
            logits.append([0] * self.vocab_size)

        return {
            PREDICTIONS: torch.tensor(matched_labels, device=outputs_device),
            PROBABILITIES: torch.tensor(probabilities, dtype=torch.float32, device=outputs_device),
            LOGITS: torch.tensor(logits, dtype=torch.float32, device=outputs_device),
        }



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

import pandas as pd

from ludwig.datasets.loaders.dataset_loader import DatasetLoader


class InsuranceLiteLoader(DatasetLoader):
    """Health Insurance Cross Sell Prediction Predict Health Insurance Owners' who will be interested in Vehicle
    Insurance https://www.kaggle.com/datasets/arashnic/imbalanced-data-practice."""

    def transform_dataframe(self, dataframe: pd.DataFrame) -> pd.DataFrame:
        df = super().transform_dataframe(dataframe)
        # Make image paths relative to dataset root directory
        df["image_path"] = df["image_path"].apply(
            lambda x: os.path.join("Fast_Furious_Insured", "trainImages", os.path.basename(x))
        )
        return df



from collections import defaultdict

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import MODEL_ECD, MODEL_GBM, MODEL_LLM
from ludwig.schema import utils as schema_utils
from ludwig.utils.registry import Registry

input_config_registries = defaultdict(Registry)
output_config_registries = defaultdict(Registry)

ecd_input_config_registry = input_config_registries[MODEL_ECD]
gbm_input_config_registry = input_config_registries[MODEL_GBM]
llm_input_config_registry = input_config_registries[MODEL_LLM]

ecd_output_config_registry = output_config_registries[MODEL_ECD]
gbm_output_config_registry = output_config_registries[MODEL_GBM]
llm_output_config_registry = output_config_registries[MODEL_LLM]

input_mixin_registry = Registry()
output_mixin_registry = Registry()

"""
As of Ludwig v0.7, ECD models support the full range of feature parameters available in Ludwig, so any feature schema
can be registered into it. See `BinaryDefaultsConfig` for an example.
"""
ecd_defaults_config_registry = Registry()

"""
As of Ludwig v0.7, GBM models only support certain feature types and those features may only contain preprocessing
parameters (in comparison, ECD features can specify encoders and other parameters). This is why the two model types have
separate defaults registries. See `BinaryInputFeatureConfigMixin` for an example of a schema pattern that is designed to
be registered by this registry (whereas, conversely, `BinaryDefaultsConfig` is an example of one to be registered with
the ECD defaults registry).
"""
gbm_defaults_config_registry = Registry()

llm_defaults_config_registry = Registry()


def input_config_registry(model_type: str) -> Registry:
    return input_config_registries[model_type]


def output_config_registry(model_type: str) -> Registry:
    return output_config_registries[model_type]


@DeveloperAPI
def get_input_feature_cls(model_type: str, name: str):
    # TODO(travis): not needed once we remove existing model config implementation
    return input_config_registries[model_type][name]


@DeveloperAPI
def get_output_feature_cls(model_type: str, name: str):
    # TODO(ksbrar): What is this?
    return output_config_registries[model_type][name]


@DeveloperAPI
def get_input_feature_jsonschema(model_type: str):
    """This function returns a JSON schema structured to only requires a `type` key and then conditionally applies
    a corresponding input feature's field constraints.

    Returns: JSON Schema
    """
    input_feature_types = sorted(list(input_config_registry(model_type).keys()))
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string", "title": "name", "description": "Name of the input feature."},
            "type": {
                "type": "string",
                "enum": input_feature_types,
                "title": "type",
                "description": "Type of the input feature",
            },
            "column": {"type": "string", "title": "column", "description": "Name of the column."},
        },
        "additionalProperties": True,
        "allOf": get_input_feature_conds(model_type),
        "required": ["name", "type"],
        "title": "input_feature",
    }

    return schema


@DeveloperAPI
def get_input_feature_conds(model_type: str):
    """This function returns a list of if-then JSON clauses for each input feature type along with their properties
    and constraints.

    Returns: List of JSON clauses
    """
    input_feature_types = sorted(list(input_config_registry(model_type).keys()))
    conds = []
    for feature_type in input_feature_types:
        schema_cls = get_input_feature_cls(model_type, feature_type)
        feature_schema = schema_utils.unload_jsonschema_from_marshmallow_class(schema_cls)
        feature_props = feature_schema["properties"]
        schema_utils.remove_duplicate_fields(feature_props)

        feature_cond = schema_utils.create_cond({"type": feature_type}, feature_props)
        conds.append(feature_cond)
    return conds


@DeveloperAPI
def get_output_feature_jsonschema(model_type: str):
    """This function returns a JSON schema structured to only requires a `type` key and then conditionally applies
    a corresponding output feature's field constraints.

    Returns: JSON Schema
    """
    output_feature_types = sorted(list(output_config_registry(model_type).keys()))
    schema = {
        "type": "object",
        "properties": {
            "name": {"type": "string", "title": "name", "description": "Name of the output feature."},
            "type": {
                "type": "string",
                "enum": output_feature_types,
                "title": "type",
                "description": "Type of the output feature",
            },
            "column": {"type": "string", "title": "column", "description": "Name of the column."},
        },
        "additionalProperties": True,
        "allOf": get_output_feature_conds(model_type),
        "required": ["name", "type"],
        "title": "output_feature",
    }

    return schema


@DeveloperAPI
def get_output_feature_conds(model_type: str):
    """This function returns a list of if-then JSON clauses for each output feature type along with their
    properties and constraints.

    Returns: List of JSON clauses
    """
    output_feature_types = sorted(list(output_config_registry(model_type).keys()))
    conds = []
    for feature_type in output_feature_types:
        schema_cls = get_output_feature_cls(model_type, feature_type)
        feature_schema = schema_utils.unload_jsonschema_from_marshmallow_class(schema_cls)
        feature_props = feature_schema["properties"]
        schema_utils.remove_duplicate_fields(feature_props)
        feature_cond = schema_utils.create_cond({"type": feature_type}, feature_props)
        conds.append(feature_cond)
    return conds



import json
import os
from typing import Optional, Type, Union
from unittest import mock

import pytest
import torch

import ludwig.schema.encoders.utils as schema_encoders_utils
from ludwig.api import LudwigModel
from ludwig.constants import ENCODER, ENCODER_OUTPUT, MODEL_ECD, NAME, TEXT, TRAINER
from ludwig.encoders import text_encoders
from ludwig.error import ConfigValidationError
from ludwig.globals import MODEL_HYPERPARAMETERS_FILE_NAME
from ludwig.schema.model_config import ModelConfig
from ludwig.utils.data_utils import load_json
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated
from tests.integration_tests.utils import (
    category_feature,
    clear_huggingface_cache,
    generate_data,
    HF_ENCODERS,
    LocalTestBackend,
    text_feature,
)

DEVICE = get_torch_device()
RANDOM_SEED = 1919


def _load_pretrained_hf_model_no_weights(
    modelClass: Type,
    pretrained_model_name_or_path: Optional[Union[str, os.PathLike]],
    **pretrained_kwargs,
):
    """Loads a HF model architecture without loading the weights."""
    from transformers import AutoConfig, AutoModel

    config = AutoConfig.from_pretrained(pretrained_model_name_or_path)
    return AutoModel.from_config(config), False


def get_mismatched_config_params(ludwig_results_dir, ludwig_model):
    saved_config_dict = load_json(os.path.join(ludwig_results_dir, "model", MODEL_HYPERPARAMETERS_FILE_NAME))
    saved_config_obj = ModelConfig.from_dict(saved_config_dict)

    mismatches = []
    for input_feature_config in saved_config_obj.input_features.to_list():
        feature_name = input_feature_config[NAME]
        encoder_config_from_file = input_feature_config[ENCODER]
        encoder_config_from_model = ludwig_model.model.input_features.get(feature_name).encoder_obj.config.to_dict()
        for k, v in encoder_config_from_model.items():
            # Skip saved_weights_in_checkpoint because this value is not yet set when the global config
            # is modified with the final encoder config.
            if k == "saved_weights_in_checkpoint":
                continue

            if encoder_config_from_file[k] != v:
                mismatch = {
                    "feature_name": feature_name,
                    "param_name": k,
                    "val_from_file": encoder_config_from_file[k],
                    "val_from_model": v,
                }
                mismatches.append(mismatch)
    return mismatches


@pytest.mark.slow
@pytest.mark.parametrize("encoder_name", HF_ENCODERS)
def test_hf_ludwig_model_e2e(tmpdir, csv_filename, encoder_name):
    """Tests HuggingFace encoders end-to-end.

    This test validates the following:
        1. Encoder config defaults are compatible with Ludwig experiments.
        2. Ludwig correctly updates the encoder config with the parameters introduced by the HF encoder.
        3. Ludwig correctly loads checkpoints containing HF encoder weights.
    """
    input_features = [
        text_feature(
            encoder={
                "vocab_size": 30,
                "min_len": 1,
                "type": encoder_name,
            }
        )
    ]
    output_features = [category_feature(decoder={"vocab_size": 2})]
    rel_path = generate_data(input_features, output_features, csv_filename)

    if encoder_name == "auto_transformer":
        # need to explciitly set the pretrained model name for auto_transformer
        input_features[0][ENCODER][
            "pretrained_model_name_or_path"
        ] = "hf-internal-testing/tiny-bert-for-token-classification"

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"train_steps": 1},
    }
    model = LudwigModel(config=config, backend=LocalTestBackend())

    with mock.patch(
        "ludwig.encoders.text_encoders.load_pretrained_hf_model_with_hub_fallback",
        side_effect=_load_pretrained_hf_model_no_weights,
    ):
        # Validates that the defaults associated with the encoder are compatible with Ludwig training.
        _, _, _, results_dir = model.experiment(dataset=rel_path, output_directory=tmpdir)

        # Validate that the saved config reflects the parameters introduced by the HF encoder.
        # This ensures that the config updates after initializing the encoder.
        mismatched_config_params = get_mismatched_config_params(results_dir, model)
        if len(mismatched_config_params) > 0:
            raise AssertionError(
                f"Config parameters mismatched with encoder parameters: "
                f"{json.dumps(mismatched_config_params, indent=4)}"
            )

        # Validate the model can be loaded.
        # This ensures that the config reflects the internal architecture of the encoder.
        LudwigModel.load(os.path.join(results_dir, "model"))
    clear_huggingface_cache()


@pytest.mark.slow
@pytest.mark.parametrize("reduce_output", [None, "last", "sum", "mean", "max", "concat"])
@pytest.mark.parametrize("encoder_name", HF_ENCODERS)
def test_hf_ludwig_model_reduce_options(tmpdir, csv_filename, encoder_name, reduce_output):
    input_features = [
        text_feature(
            preprocessing={
                "max_sequence_length": 10,
            },
            encoder={
                "vocab_size": 30,
                "min_len": 1,
                "type": encoder_name,
                "reduce_output": reduce_output,
            },
        )
    ]
    output_features = [category_feature(decoder={"vocab_size": 2})]
    rel_path = generate_data(input_features, output_features, csv_filename)

    if encoder_name == "auto_transformer":
        # need to explciitly set the pretrained model name for auto_transformer
        input_features[0][ENCODER][
            "pretrained_model_name_or_path"
        ] = "hf-internal-testing/tiny-bert-for-token-classification"

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"train_steps": 1},
    }

    try:
        ModelConfig.from_dict(config)
    except ConfigValidationError as e:
        pytest.skip(e.message)

    model = LudwigModel(
        config=config,
        backend=LocalTestBackend(),
    )

    # Validates that the defaults associated with the encoder are compatible with Ludwig training.
    with mock.patch(
        "ludwig.encoders.text_encoders.load_pretrained_hf_model_with_hub_fallback",
        side_effect=_load_pretrained_hf_model_no_weights,
    ):
        model.train(
            dataset=rel_path,
            output_directory=tmpdir,
            skip_save_training_description=True,
            skip_save_training_statistics=True,
            skip_save_model=True,
            skip_save_progress=True,
            skip_save_log=True,
            skip_save_processed_input=True,
        )

    clear_huggingface_cache()


@pytest.mark.parametrize(
    "pretrained_model_name_or_path",
    [
        "hf-internal-testing/tiny-random-bloom",
        "hf-internal-testing/tiny-random-OPTModel",
        "hf-internal-testing/tiny-random-GPTJModel",
    ],
)
def test_hf_ludwig_model_auto_transformers(tmpdir, csv_filename, pretrained_model_name_or_path):
    """Tests different AutoModel types to ensure our wrapper handles them correctly.

    This is needed because different PretrainedModel implemetnations have different input / output signatures.
    """
    input_features = [
        text_feature(
            preprocessing={
                "max_sequence_length": 10,
            },
            encoder={
                "vocab_size": 30,
                "min_len": 1,
                "type": "auto_transformer",
                "pretrained_model_name_or_path": pretrained_model_name_or_path,
            },
        )
    ]
    output_features = [category_feature(decoder={"vocab_size": 2})]
    rel_path = generate_data(input_features, output_features, csv_filename)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"train_steps": 1},
    }
    model = LudwigModel(config=config, backend=LocalTestBackend())

    # Validates that the defaults associated with the encoder are compatible with Ludwig training.
    with mock.patch(
        "ludwig.encoders.text_encoders.load_pretrained_hf_model_with_hub_fallback",
        side_effect=_load_pretrained_hf_model_no_weights,
    ):
        model.train(dataset=rel_path, output_directory=tmpdir)


@pytest.mark.parametrize("trainable", [True, False])
def test_distilbert_param_updates(trainable: bool):
    max_sequence_length = 20
    distil_bert_encoder = text_encoders.DistilBERTEncoder(
        use_pretrained=False,
        max_sequence_length=max_sequence_length,
        trainable=trainable,
    )

    # send a random input through the model with its initial weights
    inputs = torch.rand((2, max_sequence_length)).type(distil_bert_encoder.input_dtype)
    outputs = distil_bert_encoder(inputs)

    # perform a backward pass to update the model params
    target = torch.randn(outputs[ENCODER_OUTPUT].shape)
    check_module_parameters_updated(distil_bert_encoder, (inputs,), target)

    # send the same input through the model again. should be different if trainable, else the same
    outputs2 = distil_bert_encoder(inputs)

    encoder_output1 = outputs[ENCODER_OUTPUT]
    encoder_output2 = outputs2[ENCODER_OUTPUT]

    if trainable:
        # Outputs should be different if the model was updated
        assert not torch.equal(encoder_output1, encoder_output2)
    else:
        # Outputs should be the same if the model wasn't updated
        assert torch.equal(encoder_output1, encoder_output2)


@pytest.mark.parametrize("encoder_name", HF_ENCODERS)
def test_encoder_names_constant_synced_with_schema(encoder_name):
    """Ensures that each value in the HF_ENCODERS constant is represented by an equivalent schema object."""
    schema_encoders_utils.get_encoder_cls(MODEL_ECD, TEXT, encoder_name)


@pytest.mark.parametrize("vocab_size", [20])
def test_tfidf_encoder(vocab_size: int):
    # make repeatable
    torch.manual_seed(RANDOM_SEED)

    batch_size = 10
    sequence_length = 32
    vocab = [str(i) for i in range(1, vocab_size + 1)]
    str2idf = {s: 1 for s in vocab}
    text_encoder = text_encoders.TfIdfEncoder(
        max_sequence_length=sequence_length,
        str2idf=str2idf,
        vocab=vocab,
        vocab_size=vocab_size,
    ).to(DEVICE)

    assert len(text_encoder.output_shape) == 1
    assert text_encoder.output_shape[0] == vocab_size
    assert len(list(text_encoder.parameters())) == 0

    inputs = torch.randint(2, (batch_size, sequence_length)).to(DEVICE)
    outputs = text_encoder(inputs)
    assert outputs[ENCODER_OUTPUT].shape[1:] == text_encoder.output_shape


def test_hf_auto_transformer_use_pretrained():
    """This test ensures that use_pretrained is always True when using the auto_transformer text encoder even if a
    user explicitly sets it to False."""
    config = {
        "input_features": [
            text_feature(
                encoder={
                    "type": "auto_transformer",
                    "use_pretrained": False,
                    "pretrained_model_name_or_path": "hf-internal-testing/tiny-random-bloom",
                },
            )
        ],
        "output_features": [category_feature(decoder={"vocab_size": 2})],
    }

    model = LudwigModel(config=config, backend=LocalTestBackend())
    assert model.config_obj.input_features[0].encoder.use_pretrained



import contextlib
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TYPE_CHECKING, Union

import torch
from torch import nn
from torch.optim import Optimizer

from ludwig.modules.optimization_modules import create_optimizer
from ludwig.utils.torch_utils import get_torch_device

if TYPE_CHECKING:
    from ray.train.backend import BackendConfig
    from ray.train.data_parallel_trainer import DataParallelTrainer

    from ludwig.models.base import BaseModel
    from ludwig.modules.lr_scheduler import LRScheduler
    from ludwig.schema.trainer import ECDTrainerConfig
    from ludwig.utils.checkpoint_utils import Checkpoint


class DistributedStrategy(ABC):
    """Interface that wraps a distributed training framework (Horovod, DDP).

    Distributed strategies modify the model and/or optimizer to coordinate gradient updates among multiple workers
    running in parallel. In most cases, these are using collective communication libraries pass messages between
    processes.
    """

    @abstractmethod
    def prepare(
        self,
        model: nn.Module,
        trainer_config: "ECDTrainerConfig",
        base_learning_rate: float,
    ) -> Tuple[nn.Module, Optimizer]:
        """Modifies the model to support distributed training and creates the optimizer.

        Args:
            model: The model to wrap for distributed training.
            trainer_config: The trainer configuration, which includes optimizer params.
            base_learning_rate: The base learning rate to init the optimizer, which may be scaled by the strategy.

        Returns:
            A tuple of the wrapped model and the optimizer.
        """
        pass

    def prepare_for_inference(self, model: nn.Module) -> nn.Module:
        return model

    def to_device(self, model: "BaseModel", device: Optional[torch.device] = None) -> nn.Module:
        return model.to_device(device if device is not None else get_torch_device())

    def backward(self, loss: torch.Tensor, model: nn.Module):
        loss.backward()

    def step(self, optimizer: Optimizer, *args, **kwargs):
        optimizer.step(*args, **kwargs)

    def zero_grad(self, optimizer: Optimizer):
        optimizer.zero_grad()

    def set_batch_size(self, model: nn.Module, batch_size: int):
        pass

    @abstractmethod
    def size(self) -> int:
        pass

    @abstractmethod
    def rank(self) -> int:
        pass

    @abstractmethod
    def local_size(self) -> int:
        pass

    @abstractmethod
    def local_rank(self) -> int:
        pass

    def is_coordinator(self) -> bool:
        return self.rank() == 0

    @abstractmethod
    def barrier(self):
        pass

    @abstractmethod
    def allreduce(self, t: torch.Tensor) -> torch.Tensor:
        pass

    @abstractmethod
    def broadcast(self, t: torch.Tensor) -> torch.Tensor:
        pass

    @abstractmethod
    def sync_model(self, model: nn.Module):
        pass

    @abstractmethod
    def sync_optimizer(self, optimizer: Optimizer):
        pass

    @abstractmethod
    def broadcast_object(self, v: Any, name: Optional[str] = None) -> Any:
        pass

    @abstractmethod
    def wait_optimizer_synced(self, optimizer: Optimizer):
        pass

    @abstractmethod
    @contextlib.contextmanager
    def prepare_model_update(self, model: nn.Module, should_step: bool):
        pass

    @abstractmethod
    @contextlib.contextmanager
    def prepare_optimizer_update(self, optimizer: Optimizer):
        pass

    @classmethod
    @abstractmethod
    def is_available(cls) -> bool:
        pass

    @classmethod
    @abstractmethod
    def gather_all_tensors_fn(cls) -> Optional[Callable]:
        pass

    @classmethod
    @abstractmethod
    def get_ray_trainer_backend(cls, **kwargs) -> Optional[Any]:
        pass

    @classmethod
    @abstractmethod
    def get_trainer_cls(cls, backend_config: "BackendConfig") -> Tuple[Type["DataParallelTrainer"], Dict[str, Any]]:
        pass

    @abstractmethod
    def shutdown(self):
        pass

    def return_first(self, fn: Callable) -> Callable:
        """Wraps function so results are only returned by the first (coordinator) rank.

        The purpose of this function is to reduce network overhead.
        """

        def wrapped(*args, **kwargs):
            res = fn(*args, **kwargs)
            return res if self.rank() == 0 else None

        return wrapped

    def allow_gradient_accumulation(self) -> bool:
        return True

    def allow_mixed_precision(self) -> bool:
        return True

    def allow_clip_gradients(self) -> bool:
        return True

    def prepare_before_load(self) -> bool:
        """True if we need to call `prepare` again before loading a checkpoint."""
        return False

    @classmethod
    def is_model_parallel(cls) -> bool:
        return False

    def create_checkpoint_handle(
        self,
        dist_model: nn.Module,
        model: nn.Module,
        optimizer: Optional[Optimizer] = None,
        scheduler: Optional["LRScheduler"] = None,
    ) -> "Checkpoint":
        from ludwig.utils.checkpoint_utils import MultiNodeCheckpoint

        return MultiNodeCheckpoint(self, model, optimizer, scheduler)

    @classmethod
    def extract_model_for_serialization(cls, model: nn.Module) -> Union[nn.Module, Tuple[nn.Module, List[Dict]]]:
        return model

    @classmethod
    def replace_model_from_serialization(cls, state: Union[nn.Module, Tuple[nn.Module, List[Dict]]]) -> nn.Module:
        assert isinstance(state, nn.Module)
        return state


class LocalStrategy(DistributedStrategy):
    def prepare(
        self,
        model: nn.Module,
        trainer_config: "ECDTrainerConfig",
        base_learning_rate: float,
    ) -> Tuple[nn.Module, Optimizer]:
        return model, create_optimizer(model, trainer_config.optimizer, base_learning_rate)

    def size(self) -> int:
        return 1

    def rank(self) -> int:
        return 0

    def local_size(self) -> int:
        return 0

    def local_rank(self) -> int:
        return 0

    def barrier(self):
        pass

    def allreduce(self, t: torch.Tensor) -> torch.Tensor:
        return t

    def broadcast(self, t: torch.Tensor) -> torch.Tensor:
        return t

    def sync_model(self, model: nn.Module):
        pass

    def sync_optimizer(self, optimizer: Optimizer):
        pass

    def broadcast_object(self, v: Any, name: Optional[str] = None) -> Any:
        return v

    def wait_optimizer_synced(self, optimizer: Optimizer):
        pass

    @contextlib.contextmanager
    def prepare_model_update(self, model: nn.Module, should_step: bool):
        yield

    @contextlib.contextmanager
    def prepare_optimizer_update(self, optimizer: Optimizer):
        yield

    @classmethod
    def is_available(cls) -> bool:
        # While this strategy is always an option, it is not "distributed" which is the meaning of availability
        # in this context.
        return False

    @classmethod
    def gather_all_tensors_fn(cls) -> Optional[Callable]:
        return None

    @classmethod
    def get_ray_trainer_backend(cls, **kwargs) -> Optional[Any]:
        return None

    @classmethod
    def get_trainer_cls(cls, backend_config: "BackendConfig") -> Tuple[Type["DataParallelTrainer"], Dict[str, Any]]:
        raise ValueError("Cannot construct a trainer from a local strategy.")

    def shutdown(self):
        pass



import os

import pytest

from ludwig.constants import (
    BATCH_SIZE,
    DECODER,
    ENCODER,
    EPOCHS,
    INPUT_FEATURES,
    OUTPUT_FEATURES,
    SEQUENCE,
    TEXT,
    TRAINER,
    TYPE,
)
from tests.integration_tests.utils import (
    create_data_set_to_use,
    generate_data,
    RAY_BACKEND_CONFIG,
    sequence_feature,
    text_feature,
    train_with_backend,
)

pytestmark = pytest.mark.integration_tests_c


@pytest.mark.slow
@pytest.mark.parametrize("feature_type,feature_gen", [(TEXT, text_feature), (SEQUENCE, sequence_feature)])
@pytest.mark.parametrize("decoder_type", ["generator", "tagger"])
@pytest.mark.distributed
def test_sequence_decoder_predictions(tmpdir, csv_filename, ray_cluster_2cpu, feature_type, feature_gen, decoder_type):
    """Test that sequence decoders return the correct successfully predict."""
    input_feature = feature_gen()
    output_feature = feature_gen(output_feature=True)

    input_feature[ENCODER] = {TYPE: "embed", "reduce_output": None}
    output_feature[DECODER] = {TYPE: decoder_type}

    dataset_path = generate_data(
        input_features=[input_feature],
        output_features=[output_feature],
        filename=os.path.join(tmpdir, csv_filename),
    )
    dataset_path = create_data_set_to_use("csv", dataset_path)

    config = {INPUT_FEATURES: [input_feature], TRAINER: {EPOCHS: 1, BATCH_SIZE: 4}}

    # Ensure that the decoder outputs the correct predictions through both the default and feature-specific configs.
    config[OUTPUT_FEATURES] = [output_feature]

    # Test with decoder in output feature config
    train_with_backend(RAY_BACKEND_CONFIG, config=config, dataset=dataset_path)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Dict, List, TYPE_CHECKING, Union

import numpy as np
import torch

from ludwig.constants import COLUMN, HIDDEN, LOGITS, NAME, PREDICTIONS, PROC_COLUMN, TIMESERIES
from ludwig.features.base_feature import BaseFeatureMixin, OutputFeature, PredictModule
from ludwig.features.sequence_feature import SequenceInputFeature
from ludwig.features.vector_feature import _VectorPostprocessing, _VectorPredict
from ludwig.schema.features.timeseries_feature import TimeseriesInputFeatureConfig, TimeseriesOutputFeatureConfig
from ludwig.types import FeatureMetadataDict, PreprocessingConfigDict, TrainingSetMetadataDict
from ludwig.utils.tokenizers import get_tokenizer_from_registry, TORCHSCRIPT_COMPATIBLE_TOKENIZERS
from ludwig.utils.types import Series, TorchscriptPreprocessingInput

if TYPE_CHECKING:
    from ludwig.backend.base import Backend

logger = logging.getLogger(__name__)


def create_time_delay_embedding(
    series: Series, window_size: int, horizon: int, padding_value: int, backend: "Backend"
) -> Series:
    """Time delay embedding from:

    https://towardsdatascience.com/machine-learning-for-forecasting-transformations-and-feature-extraction-bbbea9de0ac2

    Args:
        series: Column-major timeseries data.
        window_size: Size of the lookback sliding window for timeseries inputs.
        horizon: Size of the forward-looking horizon for timeseries outputs.
        padding_value: Value to pad out the window when there is not enough data around the observation.

    Returns:
        A column of timeseries window arrays in row-major format for training.
    """
    # Replace default fill value of "" with nan as we will be assuming numeric values here
    series = series.replace("", np.nan)

    # Create the list of shifts we want to perform over the series.
    # For backwards looking shifts, we want to include the current element, while for forward looking shifts we do not.
    # Example:
    #   window_size=3, horizon=0 --> shift_offsets=[2, 1, 0]
    #   window_size=0, horizon=2 --> shift_offsets=[-1, -2]
    shift_offsets = list(range(window_size - 1, -(horizon + 1), -1))
    shifts = [series.shift(i) for i in shift_offsets]
    df = backend.df_engine.df_lib.concat(shifts, axis=1)
    df.columns = [f"__tmp_column_{j}" for j in shift_offsets]
    return df.apply(lambda x: np.nan_to_num(np.array(x.tolist()).astype(np.float32), nan=padding_value), axis=1)


class _TimeseriesPreprocessing(torch.nn.Module):
    """Torchscript-enabled version of preprocessing done by TimeseriesFeatureMixin.add_feature_data."""

    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        if metadata["preprocessing"]["tokenizer"] not in TORCHSCRIPT_COMPATIBLE_TOKENIZERS:
            raise ValueError(
                f"{metadata['preprocessing']['tokenizer']} is not supported by torchscript. Please use "
                f"one of {TORCHSCRIPT_COMPATIBLE_TOKENIZERS}."
            )
        self.tokenizer = get_tokenizer_from_registry(metadata["preprocessing"]["tokenizer"])()
        self.padding = metadata["preprocessing"]["padding"]
        self.padding_value = float(metadata["preprocessing"]["padding_value"])
        self.max_timeseries_length = int(metadata["max_timeseries_length"])
        self.computed_fill_value = metadata["preprocessing"]["computed_fill_value"]

    def _process_str_sequence(self, sequence: List[str], limit: int) -> torch.Tensor:
        float_sequence = [float(s) for s in sequence[:limit]]
        return torch.tensor(float_sequence)

    def _nan_to_fill_value(self, v: torch.Tensor) -> torch.Tensor:
        if v.isnan().any():
            tokenized_fill_value = self.tokenizer(self.computed_fill_value)
            # refines type of sequences from Any to List[str]
            assert torch.jit.isinstance(tokenized_fill_value, List[str])
            return self._process_str_sequence(tokenized_fill_value, self.max_timeseries_length)
        return v

    def forward_list_of_tensors(self, v: List[torch.Tensor]) -> torch.Tensor:
        v = [self._nan_to_fill_value(v_i) for v_i in v]

        if self.padding == "right":
            timeseries_matrix = torch.nn.utils.rnn.pad_sequence(v, batch_first=True, padding_value=self.padding_value)
            timeseries_matrix = timeseries_matrix[:, : self.max_timeseries_length]
        else:
            reversed_timeseries = [torch.flip(v_i[: self.max_timeseries_length], dims=(0,)) for v_i in v]
            reversed_timeseries_padded = torch.nn.utils.rnn.pad_sequence(
                reversed_timeseries, batch_first=True, padding_value=self.padding_value
            )
            timeseries_matrix = torch.flip(reversed_timeseries_padded, dims=(1,))
        return timeseries_matrix

    def forward_list_of_strs(self, v: List[str]) -> torch.Tensor:
        v = [self.computed_fill_value if s == "nan" else s for s in v]

        sequences = self.tokenizer(v)
        # refines type of sequences from Any to List[List[str]]
        assert torch.jit.isinstance(sequences, List[List[str]]), "sequences is not a list of lists."

        timeseries_matrix = torch.full(
            [len(sequences), self.max_timeseries_length], self.padding_value, dtype=torch.float32
        )
        for sample_idx, str_sequence in enumerate(sequences):
            limit = min(len(str_sequence), self.max_timeseries_length)
            float_sequence = self._process_str_sequence(str_sequence, limit)
            if self.padding == "right":
                timeseries_matrix[sample_idx][:limit] = float_sequence
            else:  # if self.padding == 'left
                timeseries_matrix[sample_idx][self.max_timeseries_length - limit :] = float_sequence
        return timeseries_matrix

    def forward(self, v: TorchscriptPreprocessingInput) -> torch.Tensor:
        """Takes a list of float values and creates a padded torch.Tensor."""
        if torch.jit.isinstance(v, List[torch.Tensor]):
            return self.forward_list_of_tensors(v)
        if torch.jit.isinstance(v, List[str]):
            return self.forward_list_of_strs(v)
        raise ValueError(f"Unsupported input: {v}")


class TimeseriesFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return TIMESERIES

    @staticmethod
    def cast_column(column, backend):
        return column

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        window_size = preprocessing_parameters.get("window_size", 0) or preprocessing_parameters.get("horizon", 0)
        if window_size > 0:
            # Column-major data
            return {"max_timeseries_length": window_size}

        column = column.astype(str)
        tokenizer = get_tokenizer_from_registry(preprocessing_parameters["tokenizer"])()
        max_length = 0
        for timeseries in column:
            processed_line = tokenizer(timeseries)
            max_length = max(max_length, len(processed_line))
        max_length = min(preprocessing_parameters["timeseries_length_limit"], max_length)

        return {"max_timeseries_length": max_length}

    @staticmethod
    def build_matrix(timeseries, tokenizer_name, length_limit, padding_value, padding, backend):
        tokenizer = get_tokenizer_from_registry(tokenizer_name)()

        ts_vectors = backend.df_engine.map_objects(
            timeseries, lambda ts: np.nan_to_num(np.array(tokenizer(ts)).astype(np.float32), nan=padding_value)
        )

        max_length = backend.df_engine.compute(ts_vectors.map(len).max())
        if max_length < length_limit:
            logger.debug(f"max length of {tokenizer_name}: {max_length} < limit: {length_limit}")
        max_length = length_limit

        def pad(vector):
            padded = np.full((max_length,), padding_value, dtype=np.float32)
            limit = min(vector.shape[0], max_length)
            if padding == "right":
                padded[:limit] = vector[:limit]
            else:  # if padding == 'left
                padded[max_length - limit :] = vector[:limit]
            return padded

        return backend.df_engine.map_objects(ts_vectors, pad)

    @staticmethod
    def feature_data(column, metadata, preprocessing_parameters: PreprocessingConfigDict, backend):
        padding_value = preprocessing_parameters["padding_value"]

        window_size = preprocessing_parameters.get("window_size", 0)
        horizon = preprocessing_parameters.get("horizon", 0)
        if window_size > 0 or horizon > 0:
            # Column-major data. Convert the column into the row-major embedding
            return create_time_delay_embedding(column, window_size, horizon, padding_value, backend)

        timeseries_data = TimeseriesFeatureMixin.build_matrix(
            column,
            preprocessing_parameters["tokenizer"],
            metadata["max_timeseries_length"],
            padding_value,
            preprocessing_parameters["padding"],
            backend,
        )
        return timeseries_data

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        proc_df[feature_config[PROC_COLUMN]] = TimeseriesFeatureMixin.feature_data(
            input_df[feature_config[COLUMN]].astype(str),
            metadata[feature_config[NAME]],
            preprocessing_parameters,
            backend,
        )
        return proc_df


class TimeseriesInputFeature(TimeseriesFeatureMixin, SequenceInputFeature):
    def __init__(self, input_feature_config: TimeseriesInputFeatureConfig, encoder_obj=None, **kwargs):
        # add required sequence encoder parameters for time series
        input_feature_config.encoder.embedding_size = 1
        input_feature_config.encoder.should_embed = False

        # SequenceInputFeauture's constructor initializes the encoder.
        super().__init__(input_feature_config, encoder_obj=encoder_obj, **kwargs)

    def forward(self, inputs, mask=None):
        assert isinstance(inputs, torch.Tensor)
        assert inputs.dtype in [torch.float16, torch.float32, torch.float64]
        assert len(inputs.shape) == 2

        inputs_exp = inputs.type(torch.float32)
        encoder_output = self.encoder_obj(inputs_exp, mask=mask)

        return encoder_output

    @property
    def input_shape(self) -> torch.Size:
        return self.encoder_obj.input_shape

    @property
    def input_dtype(self):
        return torch.float32

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.encoder.input_size = feature_metadata["max_timeseries_length"]
        feature_config.encoder.max_sequence_length = feature_metadata["max_timeseries_length"]

    @staticmethod
    def get_schema_cls():
        return TimeseriesInputFeatureConfig

    @staticmethod
    def create_preproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _TimeseriesPreprocessing(metadata)


class TimeseriesOutputFeature(TimeseriesFeatureMixin, OutputFeature):
    def __init__(
        self,
        output_feature_config: Union[TimeseriesOutputFeatureConfig, Dict],
        output_features: Dict[str, OutputFeature],
        **kwargs,
    ):
        self.horizon = output_feature_config.horizon
        super().__init__(output_feature_config, output_features, **kwargs)
        output_feature_config.decoder.output_size = self.horizon

        self.decoder_obj = self.initialize_decoder(output_feature_config.decoder)
        self._setup_loss()
        self._setup_metrics()

    def logits(self, inputs, **kwargs):  # hidden
        hidden = inputs[HIDDEN]
        return self.decoder_obj(hidden)

    def loss_kwargs(self):
        return self.loss.to_dict()

    def metric_kwargs(self):
        return dict(num_outputs=self.output_shape[0])

    def create_predict_module(self) -> PredictModule:
        return _VectorPredict()

    def get_prediction_set(self):
        return {PREDICTIONS, LOGITS}

    @classmethod
    def get_output_dtype(cls):
        return torch.float32

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.horizon])

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.horizon = feature_metadata["max_timeseries_length"]

    @staticmethod
    def calculate_overall_stats(predictions, targets, train_set_metadata):
        # no overall stats, just return empty dictionary
        return {}

    def postprocess_predictions(
        self,
        result,
        metadata,
    ):
        predictions_col = f"{self.feature_name}_{PREDICTIONS}"
        if predictions_col in result:
            result[predictions_col] = result[predictions_col].map(lambda pred: pred.tolist())
        return result

    @staticmethod
    def create_postproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _VectorPostprocessing()

    @staticmethod
    def get_schema_cls():
        return TimeseriesOutputFeatureConfig



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import contextlib
import logging
import os
import shutil
import uuid
from collections import namedtuple

import pandas as pd
import pytest
import torch
import torchvision
import yaml

from ludwig.api import LudwigModel
from ludwig.backend import LOCAL_BACKEND
from ludwig.callbacks import Callback
from ludwig.constants import BATCH_SIZE, COLUMN, ENCODER, H3, NAME, PREPROCESSING, TRAINER, TYPE
from ludwig.data.concatenate_datasets import concatenate_df
from ludwig.data.dataset_synthesizer import build_synthetic_dataset_df
from ludwig.data.preprocessing import preprocess_for_training
from ludwig.encoders.registry import get_encoder_classes
from ludwig.error import ConfigValidationError
from ludwig.experiment import experiment_cli
from ludwig.predict import predict_cli
from ludwig.utils.data_utils import read_csv
from ludwig.utils.defaults import default_random_seed
from tests.integration_tests.utils import (
    audio_feature,
    bag_feature,
    binary_feature,
    category_distribution_feature,
    category_feature,
    create_data_set_to_use,
    date_feature,
    ENCODERS,
    generate_data,
    generate_output_features_with_dependencies,
    generate_output_features_with_dependencies_complex,
    h3_feature,
    image_feature,
    LocalTestBackend,
    number_feature,
    run_experiment,
    sequence_feature,
    set_feature,
    TEXT_ENCODERS,
    text_feature,
    timeseries_feature,
    vector_feature,
)

pytestmark = pytest.mark.integration_tests_d

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logging.getLogger("ludwig").setLevel(logging.INFO)


@pytest.mark.parametrize("encoder", TEXT_ENCODERS)
def test_experiment_text_feature_non_pretrained(encoder, csv_filename):
    input_features = [
        text_feature(encoder={"vocab_size": 30, "min_len": 1, "type": encoder}, preprocessing={"tokenizer": "space"})
    ]
    output_features = [category_feature(decoder={"vocab_size": 2})]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


def run_experiment_with_encoder(encoder, csv_filename):
    # Run in a subprocess to clear TF and prevent OOM
    # This also allows us to use GPU resources
    input_features = [text_feature(encoder={"vocab_size": 30, "min_len": 1, "type": encoder})]
    output_features = [category_feature(decoder={"vocab_size": 2})]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("encoder", ENCODERS)
def test_experiment_seq_seq_generator(csv_filename, encoder):
    input_features = [text_feature(encoder={"type": encoder, "reduce_output": None})]
    output_features = [text_feature(decoder={"type": "generator"}, output_feature=True)]
    rel_path = generate_data(input_features, output_features, csv_filename)

    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("encoder", ["embed", "rnn", "parallel_cnn", "stacked_parallel_cnn", "transformer"])
def test_experiment_seq_seq_tagger(csv_filename, encoder):
    input_features = [text_feature(encoder={"type": encoder, "reduce_output": None})]
    output_features = [text_feature(decoder={"type": "tagger"}, reduce_input=None)]
    rel_path = generate_data(input_features, output_features, csv_filename)

    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("encoder", ["cnnrnn", "stacked_cnn"])
def test_experiment_seq_seq_tagger_fails_for_non_length_preserving_encoders(csv_filename, encoder):
    input_features = [text_feature(encoder={"type": encoder, "reduce_output": None})]
    output_features = [text_feature(decoder={"type": "tagger"}, reduce_input=None)]
    rel_path = generate_data(input_features, output_features, csv_filename)

    with pytest.raises(ValueError):
        run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_seq_seq_model_def_file(csv_filename, yaml_filename):
    # seq-to-seq test to use config file instead of dictionary
    input_features = [text_feature(encoder={"reduce_output": None, "type": "embed"})]
    output_features = [text_feature(decoder={"vocab_size": 3, "type": "tagger"}, reduce_input=None)]

    # Save the config to a yaml file
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }
    with open(yaml_filename, "w") as yaml_out:
        yaml.safe_dump(config, yaml_out)

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(None, None, dataset=rel_path, config=yaml_filename)


def test_experiment_seq_seq_train_test_valid(tmpdir):
    # seq-to-seq test to use train, test, validation files
    input_features = [text_feature(encoder={"reduce_output": None, "type": "rnn"})]
    output_features = [text_feature(decoder={"vocab_size": 3, "type": "tagger"}, reduce_input=None)]

    train_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "train.csv"))
    test_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "test.csv"), 20)
    valdation_csv = generate_data(input_features, output_features, os.path.join(tmpdir, "val.csv"), 20)

    run_experiment(
        input_features, output_features, training_set=train_csv, test_set=test_csv, validation_set=valdation_csv
    )

    # Save intermediate output
    run_experiment(
        input_features, output_features, training_set=train_csv, test_set=test_csv, validation_set=valdation_csv
    )


@pytest.mark.parametrize("encoder", ENCODERS)
def test_experiment_multi_input_intent_classification(csv_filename, encoder):
    # Multiple inputs, Single category output
    input_features = [
        text_feature(encoder={"vocab_size": 10, "min_len": 1, "representation": "sparse"}),
        category_feature(encoder={"vocab_size": 10}),
    ]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    input_features[0][ENCODER][TYPE] = encoder
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_with_torch_module_dict_feature_name(csv_filename):
    input_features = [category_feature(name="type")]
    output_features = [category_feature(name="to", output_feature=True)]
    rel_path = generate_data(input_features, output_features, csv_filename)

    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_multiclass_with_class_weights(csv_filename):
    # Multiple inputs, Single category output
    input_features = [category_feature(encoder={"vocab_size": 10})]
    output_features = [category_feature(decoder={"vocab_size": 3}, loss={"class_weights": [0, 1, 2]})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_multilabel_with_class_weights(csv_filename):
    # Multiple inputs, Single category output
    input_features = [category_feature(encoder={"vocab_size": 10})]
    output_features = [set_feature(decoder={"vocab_size": 3}, loss={"class_weights": [0, 1, 2, 3]})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize(
    "output_features",
    [
        # baseline test case
        [
            category_feature(decoder={"reduce_input": "sum", "vocab_size": 2}),
            sequence_feature(decoder={"vocab_size": 10, "max_len": 5}),
            number_feature(),
        ],
        # use generator as decoder
        [
            category_feature(decoder={"vocab_size": 2, "reduce_input": "sum"}),
            sequence_feature(decoder={"vocab_size": 10, "max_len": 5, "type": "generator"}),
            number_feature(),
        ],
        # Generator decoder and reduce_input = None
        [
            category_feature(decoder={"vocab_size": 2, "reduce_input": "sum"}),
            sequence_feature(decoder={"max_len": 5, "type": "generator"}, reduce_input=None),
            number_feature(normalization="minmax"),
        ],
        # output features with dependencies single dependency
        generate_output_features_with_dependencies("number_feature", ["category_feature"]),
        # output features with dependencies multiple dependencies
        generate_output_features_with_dependencies("number_feature", ["category_feature", "sequence_feature"]),
        # output features with dependencies multiple dependencies
        generate_output_features_with_dependencies("sequence_feature", ["category_feature", "number_feature"]),
        # output features with dependencies
        generate_output_features_with_dependencies("category_feature", ["sequence_feature"]),
        generate_output_features_with_dependencies_complex(),
    ],
)
def test_experiment_multiple_seq_seq(csv_filename, output_features):
    input_features = [
        text_feature(encoder={"vocab_size": 100, "min_len": 1, "type": "stacked_cnn"}),
        number_feature(normalization="zscore"),
        category_feature(encoder={"vocab_size": 10, "embedding_size": 5}),
        set_feature(),
        sequence_feature(encoder={"vocab_size": 10, "max_len": 10, "type": "embed"}),
    ]
    output_features = output_features

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("skip_save_processed_input", [True, False])
@pytest.mark.parametrize("in_memory", [True, False])
@pytest.mark.parametrize("image_source", ["file", "tensor"])
@pytest.mark.parametrize("num_channels", [1, 3])
def test_basic_image_feature(num_channels, image_source, in_memory, skip_save_processed_input, tmpdir):
    # Image Inputs
    image_dest_folder = os.path.join(tmpdir, "generated_images")

    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={
                "in_memory": in_memory,
                "height": 12,
                "width": 12,
                "num_channels": num_channels,
                "num_processes": 5,
            },
            encoder={
                "type": "stacked_cnn",
                "output_size": 16,
                "num_filters": 8,
            },
        )
    ]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2})]

    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    if image_source == "file":
        # use images from file
        run_experiment(
            input_features, output_features, dataset=rel_path, skip_save_processed_input=skip_save_processed_input
        )
    else:
        # import image from file and store in dataframe as tensors.
        df = pd.read_csv(rel_path)
        image_feature_name = input_features[0]["name"]
        df[image_feature_name] = df[image_feature_name].apply(lambda x: torchvision.io.read_image(x))

        run_experiment(input_features, output_features, dataset=df, skip_save_processed_input=skip_save_processed_input)


def test_experiment_infer_image_metadata(tmpdir):
    # Image Inputs
    image_dest_folder = os.path.join(tmpdir, "generated_images")

    # Resnet encoder
    input_features = [
        image_feature(folder=image_dest_folder, encoder={"type": "stacked_cnn", "output_size": 16, "num_filters": 8}),
        text_feature(encoder={"type": "embed", "min_len": 1}),
        number_feature(normalization="zscore"),
    ]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2}), number_feature()]

    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    # remove image preprocessing section to force inferring image meta data
    input_features[0].pop("preprocessing")

    run_experiment(input_features, output_features, dataset=rel_path)


ImageParams = namedtuple("ImageTestParams", "image_encoder in_memory_flag skip_save_processed_input")


@pytest.mark.parametrize(
    "image_params",
    [
        ImageParams("stacked_cnn", True, True),
        ImageParams("stacked_cnn", False, False),
    ],
)
def test_experiment_image_inputs(image_params: ImageParams, tmpdir):
    # Image Inputs
    image_dest_folder = os.path.join(tmpdir, "generated_images")

    # Resnet encoder
    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
            encoder={"type": "resnet", "output_size": 16, "num_filters": 8},
        ),
        text_feature(encoder={"type": "embed", "min_len": 1}),
        number_feature(normalization="zscore"),
    ]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2}), number_feature()]

    input_features[0]["encoder"]["type"] = image_params.image_encoder
    input_features[0]["preprocessing"]["in_memory"] = image_params.in_memory_flag
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    run_experiment(
        input_features,
        output_features,
        dataset=rel_path,
        skip_save_processed_input=image_params.skip_save_processed_input,
    )


# Primary focus of this test is to determine if exceptions are raised for different data set formats and in_memory
# setting.


@pytest.mark.parametrize("test_in_memory", [True, False])
@pytest.mark.parametrize("test_format", ["csv", "df", "hdf5"])
@pytest.mark.parametrize("train_in_memory", [True, False])
@pytest.mark.parametrize("train_format", ["csv", "df", "hdf5"])
def test_experiment_image_dataset(train_format, train_in_memory, test_format, test_in_memory, tmpdir):
    # Image Inputs
    image_dest_folder = os.path.join(tmpdir, "generated_images")

    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={"in_memory": True, "height": 12, "width": 12, "num_channels": 3, "num_processes": 5},
            encoder={"type": "stacked_cnn", "output_size": 16, "num_filters": 8},
        ),
    ]
    output_features = [
        category_feature(decoder={"reduce_input": "sum", "vocab_size": 2}),
    ]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        "preprocessing": {},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    # create temporary name for train and test data sets
    train_csv_filename = os.path.join(tmpdir, "train_" + uuid.uuid4().hex[:10].upper() + ".csv")
    test_csv_filename = os.path.join(tmpdir, "test_" + uuid.uuid4().hex[:10].upper() + ".csv")

    # setup training data format to test
    train_data = generate_data(input_features, output_features, train_csv_filename)
    config["input_features"][0]["preprocessing"]["in_memory"] = train_in_memory
    training_set_metadata = None

    # define Ludwig model
    backend = LocalTestBackend()
    model = LudwigModel(
        config=config,
        backend=backend,
    )

    if train_format == "hdf5":
        # hdf5 format
        train_set, _, _, training_set_metadata = preprocess_for_training(
            model.config,
            dataset=train_data,
            backend=backend,
        )
        train_dataset_to_use = train_set.data_hdf5_fp
    else:
        train_dataset_to_use = create_data_set_to_use(train_format, train_data)

    model.train(dataset=train_dataset_to_use, training_set_metadata=training_set_metadata)

    model.config_obj.input_features.to_list()[0]["preprocessing"]["in_memory"] = test_in_memory

    # setup test data format to test
    test_data = generate_data(input_features, output_features, test_csv_filename)

    if test_format == "hdf5":
        # hdf5 format
        # create hdf5 data set
        _, test_set, _, training_set_metadata_for_test = preprocess_for_training(
            model.config,
            dataset=test_data,
            backend=backend,
        )
        test_dataset_to_use = test_set.data_hdf5_fp
    else:
        test_dataset_to_use = create_data_set_to_use(test_format, test_data)

    # run functions with the specified data format
    model.evaluate(dataset=test_dataset_to_use)
    model.predict(dataset=test_dataset_to_use)


DATA_FORMATS_TO_TEST = [
    "csv",
    "df",
    "dict",
    "excel",
    "feather",
    "fwf",
    "hdf5",
    "html",
    "json",
    "jsonl",
    "parquet",
    "pickle",
    "stata",
    "tsv",
]


@pytest.mark.parametrize("data_format", DATA_FORMATS_TO_TEST)
def test_experiment_dataset_formats(data_format, csv_filename):
    # primary focus of this test is to determine if exceptions are
    # raised for different data set formats and in_memory setting

    input_features = [number_feature(), category_feature()]
    output_features = [category_feature(output_feature=True), number_feature()]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        "preprocessing": {},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    # setup training data format to test
    raw_data = generate_data(input_features, output_features, csv_filename)

    training_set_metadata = None

    # define Ludwig model
    model = LudwigModel(config=config)

    if data_format == "hdf5":
        # hdf5 format
        training_set, _, _, training_set_metadata = preprocess_for_training(model.config, dataset=raw_data)
        dataset_to_use = training_set.data_hdf5_fp
    else:
        dataset_to_use = create_data_set_to_use(data_format, raw_data)

    model.train(dataset=dataset_to_use, training_set_metadata=training_set_metadata, random_seed=default_random_seed)

    # # run functions with the specified data format
    model.evaluate(dataset=dataset_to_use)
    model.predict(dataset=dataset_to_use)


def test_experiment_audio_inputs(tmpdir):
    # Audio Inputs
    audio_dest_folder = os.path.join(tmpdir, "generated_audio")

    input_features = [audio_feature(folder=audio_dest_folder)]
    output_features = [binary_feature()]

    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_tied_weights(csv_filename):
    # Single sequence input, single category output
    input_features = [
        text_feature(name="text_feature1", encoder={"min_len": 1, "type": "cnnrnn", "reduce_output": "sum"}),
        text_feature(
            name="text_feature2", encoder={"min_len": 1, "type": "cnnrnn", "reduce_output": "sum"}, tied="text_feature1"
        ),
    ]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    for encoder in ENCODERS:
        input_features[0][ENCODER][TYPE] = encoder
        input_features[1][ENCODER][TYPE] = encoder
        run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_tied_weights_sequence_combiner(csv_filename):
    """Tests that tied weights work with sequence combiners if `sequence_length` is provided.

    Addresses https://github.com/ludwig-ai/ludwig/issues/3220
    """
    input_features = [
        text_feature(
            name="feature1",
            encoder={
                "max_len": 5,
                "reduce_output": None,
            },
            preprocessing={"sequence_length": 10},
        ),
        text_feature(
            name="feature2",
            encoder={
                "max_len": 3,
                "reduce_output": None,
            },
            preprocessing={"sequence_length": 10},
            tied="feature1",
        ),
    ]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2})]
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "sequence"},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(config=config, dataset=rel_path)


@pytest.mark.parametrize("enc_cell_type", ["lstm", "rnn", "gru"])
@pytest.mark.parametrize("attention", [False, True])
def test_sequence_tagger(enc_cell_type, attention, csv_filename):
    # Define input and output features
    input_features = [
        sequence_feature(encoder={"max_len": 10, "type": "rnn", "cell_type": enc_cell_type, "reduce_output": None})
    ]
    output_features = [
        sequence_feature(decoder={"max_len": 10, "type": "tagger", "attention": attention}, reduce_input=None)
    ]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    # run the experiment
    run_experiment(input_features, output_features, dataset=rel_path)


def test_sequence_tagger_text(csv_filename):
    # Define input and output features
    input_features = [text_feature(encoder={"max_len": 10, "type": "rnn", "reduce_output": None})]
    output_features = [
        sequence_feature(
            decoder={"max_len": 10, "type": "tagger"},
            reduce_input=None,
        )
    ]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    # run the experiment
    run_experiment(input_features, output_features, dataset=rel_path)


"""
@pytest.mark.distributed
def test_sequence_tagger_text_ray(csv_filename, ray_cluster_2cpu):
    # Define input and output features
    input_features = [text_feature(encoder={"max_len": 10, "type": "rnn", "reduce_output": None})]
    output_features = [
        sequence_feature(
            decoder={"max_len": 10, "type": "tagger"},
            reduce_input=None,
        )
    ]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    # run the experiment
    run_experiment(input_features, output_features, dataset=rel_path, backend="ray")
"""


def test_experiment_sequence_combiner_with_reduction_fails(csv_filename):
    config = {
        "input_features": [
            sequence_feature(
                name="seq1",
                encoder={
                    "min_len": 5,
                    "max_len": 5,
                    "type": "embed",
                    "cell_type": "lstm",
                    "reduce_output": "sum",
                },
            ),
            sequence_feature(
                name="seq2",
                encoder={
                    "min_len": 5,
                    "max_len": 5,
                    "type": "embed",
                    "cell_type": "lstm",
                    "reduce_output": "sum",
                },
            ),
            category_feature(encoder={"vocab_size": 5}),
        ],
        "output_features": [category_feature(decoder={"reduce_input": "sum", "vocab_size": 5})],
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
        "combiner": {
            "type": "sequence",
            "encoder": {"type": "rnn"},
            "main_sequence_feature": "seq1",
            "reduce_output": None,
        },
    }

    # Generate test data
    rel_path = generate_data(config["input_features"], config["output_features"], csv_filename)

    # Encoding sequence features with 'embed' should fail with SequenceConcatCombiner, since at least one sequence
    # feature should be rank 3.
    with pytest.raises(TypeError):
        run_experiment(config=config, dataset=rel_path)


@pytest.mark.parametrize("sequence_encoder", ENCODERS[1:])
def test_experiment_sequence_combiner(sequence_encoder, csv_filename):
    config = {
        "input_features": [
            sequence_feature(
                name="seq1",
                encoder={
                    "min_len": 5,
                    "max_len": 5,
                    "type": sequence_encoder,
                    "cell_type": "lstm",
                    "reduce_output": None,
                },
            ),
            sequence_feature(
                name="seq2",
                encoder={
                    "min_len": 5,
                    "max_len": 5,
                    "type": sequence_encoder,
                    "cell_type": "lstm",
                    "reduce_output": None,
                },
            ),
            category_feature(vocab_size=5),
        ],
        "output_features": [category_feature(decoder={"reduce_input": "sum", "vocab_size": 5})],
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
        "combiner": {
            "type": "sequence",
            "encoder": {"type": "rnn"},
            "main_sequence_feature": "seq1",
            "reduce_output": None,
        },
    }

    # Generate test data
    rel_path = generate_data(config["input_features"], config["output_features"], csv_filename)

    run_experiment(config=config, dataset=rel_path)


def test_experiment_model_resume(tmpdir):
    # Single sequence input, single category output
    # Tests saving a model file, loading it to rerun training and predict
    input_features = [sequence_feature(encoder={"type": "rnn", "reduce_output": "sum"})]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2})]
    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    _, _, _, _, output_dir = experiment_cli(config, dataset=rel_path, output_directory=tmpdir)

    experiment_cli(config, dataset=rel_path, model_resume_path=output_dir)

    predict_cli(os.path.join(output_dir, "model"), dataset=rel_path)
    shutil.rmtree(output_dir, ignore_errors=True)


@pytest.mark.slow
@pytest.mark.parametrize(
    "dist_strategy",
    [
        pytest.param("ddp", id="ddp", marks=pytest.mark.distributed),
        pytest.param("horovod", id="horovod", marks=[pytest.mark.distributed, pytest.mark.horovod]),
    ],
)
def test_experiment_model_resume_distributed(tmpdir, dist_strategy, ray_cluster_4cpu):
    _run_experiment_model_resume_distributed(tmpdir, dist_strategy)


@pytest.mark.skipif(torch.cuda.device_count() == 0, reason="test requires at least 1 gpu")
@pytest.mark.skipif(not torch.cuda.is_available(), reason="test requires gpu support")
@pytest.mark.parametrize(
    "dist_strategy",
    [
        pytest.param("deepspeed", id="deepspeed", marks=pytest.mark.distributed),
    ],
)
def test_experiment_model_resume_distributed_gpu(tmpdir, dist_strategy, ray_cluster_4cpu):
    _run_experiment_model_resume_distributed(tmpdir, dist_strategy)


def _run_experiment_model_resume_distributed(tmpdir, dist_strategy):
    # Single sequence input, single category output
    # Tests saving a model file, loading it to rerun training and predict
    input_features = [number_feature()]
    output_features = [category_feature(output_feature=True)]
    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 8},
        TRAINER: {"epochs": 1, BATCH_SIZE: 128},
        "backend": {"type": "ray", "trainer": {"strategy": dist_strategy, "num_workers": 2}},
    }

    _, _, _, _, output_dir = experiment_cli(config, dataset=rel_path, output_directory=os.path.join(tmpdir, "results1"))

    experiment_cli(
        config, dataset=rel_path, model_resume_path=output_dir, output_directory=os.path.join(tmpdir, "results2")
    )

    predict_cli(os.path.join(output_dir, "model"), dataset=rel_path, output_directory=os.path.join(tmpdir, "results3"))


@pytest.mark.parametrize(
    "missing_file",
    ["training_progress.json", "training_checkpoints"],
    ids=["training_progress", "training_checkpoints"],
)
def test_experiment_model_resume_missing_file(tmpdir, missing_file):
    # Single sequence input, single category output
    # Tests saving a model file, loading it to rerun training and predict
    input_features = [sequence_feature(encoder={"type": "rnn", "reduce_output": "sum"})]
    output_features = [category_feature(decoder={"reduce_input": "sum", "vocab_size": 2})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    _, _, _, _, output_dir = experiment_cli(config, dataset=rel_path, output_directory=tmpdir)

    try:
        # Remove file to simulate failure during first epoch of training which prevents
        # training_checkpoints to be empty and training_progress.json to not be created
        missing_file_path = os.path.join(output_dir, "model", missing_file)
        if missing_file == "training_progress.json":
            os.remove(missing_file_path)
        else:
            shutil.rmtree(missing_file_path)
    finally:
        # Training should start a fresh model training run without any errors
        experiment_cli(config, dataset=rel_path, model_resume_path=output_dir)

    predict_cli(os.path.join(output_dir, "model"), dataset=rel_path)
    shutil.rmtree(output_dir, ignore_errors=True)


@pytest.mark.slow
@pytest.mark.distributed
def test_experiment_model_resume_before_1st_epoch_distributed(tmpdir, ray_cluster_4cpu):
    # Single sequence input, single category output
    # Tests saving a model file, loading it to rerun training and predict
    input_features = [number_feature()]
    output_features = [category_feature(output_feature=True)]
    # Generate test data
    training_set = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 8},
        TRAINER: {"train_steps": 1, BATCH_SIZE: 128},
        "backend": {"type": "ray", "trainer": {"strategy": "ddp", "num_workers": 2}},
    }

    class InducedFailureCallback(Callback):
        """Class that defines the methods necessary to hook into process."""

        def on_resume_training(self, is_coordinator):
            if is_coordinator:
                raise RuntimeError("Induced failure")

    class NoFailureCallback(Callback):
        """Class that defines the methods necessary to hook into process."""

        def on_resume_training(self, is_coordinator):
            pass

    try:
        # Define Ludwig model object that drive model training
        model = LudwigModel(config=config, logging_level=logging.INFO, callbacks=[InducedFailureCallback()])
        model.train(
            dataset=training_set,
            experiment_name="simple_experiment",
            model_name="simple_model_incomplete",
            skip_save_processed_input=True,
            output_directory=os.path.join(tmpdir, "results1"),
        )
    except RuntimeError:
        model = LudwigModel(config=config, logging_level=logging.INFO, callbacks=[NoFailureCallback()])
        model.train(
            dataset=training_set,
            skip_save_processed_input=True,
            model_resume_path=os.path.join(tmpdir, "results1"),
        )


@pytest.mark.slow
@pytest.mark.distributed
def test_tabnet_with_batch_size_1(tmpdir, ray_cluster_4cpu):
    input_features = [number_feature()]
    output_features = [category_feature(output_feature=True)]
    training_set = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "tabnet"},
        TRAINER: {"train_steps": 1, BATCH_SIZE: 1},
        "backend": {"type": "ray", "trainer": {"strategy": "ddp", "num_workers": 2}},
    }
    model = LudwigModel(config=config, logging_level=logging.INFO)
    model.train(
        dataset=training_set,
        skip_save_training_description=True,
        skip_save_training_statistics=True,
        skip_save_model=True,
        skip_save_progress=True,
        skip_save_log=True,
        skip_save_processed_input=True,
    )


def test_experiment_various_feature_types(csv_filename):
    input_features = [binary_feature(), bag_feature()]
    output_features = [set_feature(decoder={"max_len": 3, "vocab_size": 5})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_timeseries(csv_filename):
    input_features = [timeseries_feature()]
    output_features = [binary_feature()]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    input_features[0][ENCODER][TYPE] = "transformer"
    run_experiment(input_features, output_features, dataset=rel_path)


def test_visual_question_answering(tmpdir):
    image_dest_folder = os.path.join(tmpdir, "generated_images")
    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={"in_memory": True, "height": 32, "width": 32, "num_channels": 3, "num_processes": 5},
            encoder={
                "type": "stacked_cnn",
            },
        ),
        text_feature(encoder={"type": "embed", "min_len": 1}),
    ]
    output_features = [sequence_feature(decoder={"type": "generator", "cell_type": "lstm"})]
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset.csv"))

    run_experiment(input_features, output_features, dataset=rel_path)


def test_image_resizing_num_channel_handling(tmpdir):
    """This test creates two image datasets with 3 channels and 1 channel. The combination of this data is used to
    train a model. This checks the cases where the user may or may not specify a number of channels in the config.

    :param csv_filename:
    :return:
    """
    # Image Inputs
    image_dest_folder = os.path.join(tmpdir, "generated_images")

    # Resnet encoder
    input_features = [
        image_feature(
            folder=image_dest_folder,
            preprocessing={"in_memory": True, "height": 32, "width": 32, "num_channels": 3, "num_processes": 5},
            encoder={
                "type": "stacked_cnn",
            },
        ),
        text_feature(encoder={"type": "embed", "min_len": 1}),
        number_feature(normalization="minmax"),
    ]
    output_features = [binary_feature(), number_feature()]
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset1.csv"), num_examples=50)

    df1 = read_csv(rel_path)

    input_features[0]["preprocessing"]["num_channels"] = 1
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, "dataset2.csv"), num_examples=50)
    df2 = read_csv(rel_path)

    df = concatenate_df(df1, df2, None, LOCAL_BACKEND)
    df.to_csv(rel_path, index=False)

    # Here the user specifies number of channels. Exception shouldn't be thrown
    run_experiment(input_features, output_features, dataset=rel_path)

    del input_features[0]["preprocessing"]["num_channels"]

    # User doesn't specify num channels, but num channels is inferred. Exception shouldn't be thrown
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("encoder", ["wave", "embed"])
def test_experiment_date(encoder, csv_filename):
    input_features = [date_feature()]
    output_features = [category_feature(decoder={"vocab_size": 2})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    input_features[0][ENCODER] = {TYPE: encoder}
    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("encoder", get_encoder_classes(H3).keys())
def test_experiment_h3(encoder, csv_filename):
    input_features = [h3_feature()]
    output_features = [binary_feature()]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    input_features[0][ENCODER] = {TYPE: encoder}
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_vector_feature(csv_filename):
    input_features = [vector_feature()]
    output_features = [binary_feature()]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_vector_feature_infer_size(csv_filename):
    input_features = [vector_feature()]
    output_features = [vector_feature()]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    # Unset vector_size so it needs to be inferred
    del input_features[0][PREPROCESSING]
    del output_features[0][PREPROCESSING]

    run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("encoder", ["parallel_cnn", "dense", "passthrough"])
def test_forecasting_row_major(csv_filename, encoder):
    input_features = [timeseries_feature(encoder={"type": encoder})]
    output_features = [timeseries_feature(decoder={"type": "projector"})]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14, "flatten_inputs": True},
        TRAINER: {"epochs": 2, BATCH_SIZE: 128},
    }

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, config=config, dataset=rel_path)


def test_forecasting_column_major(csv_filename):
    input_feature = timeseries_feature(preprocessing={"window_size": 3})
    input_features = [input_feature]

    # Ensure output feature has the same column and the input feature
    output_feature = timeseries_feature(
        name=input_feature[COLUMN], preprocessing={"horizon": 2}, decoder={"type": "projector"}
    )
    output_feature[NAME] = f"{input_feature[NAME]}_out"
    output_features = [output_feature]

    # Generate test data in column-major format. This is just a dataframe of numbers with the same column name
    # as expected by the timeseries input feature
    column_major_feature = number_feature(name=input_feature[COLUMN])
    csv_filename = generate_data([column_major_feature], [], csv_filename)

    input_df = pd.read_csv(csv_filename)

    model, eval_stats, train_stats, preprocessed_data, output_directory = run_experiment(
        input_features, output_features, dataset=csv_filename
    )
    train_set, val_set, test_set, _ = preprocessed_data

    print(input_df)
    # print(train_set.to_df())

    horizon_df = model.forecast(input_df, horizon=5)
    print(horizon_df)


@pytest.mark.parametrize("reduce_output", [("sum"), (None)], ids=["sum", "none"])
def test_experiment_text_output_feature_with_tagger_decoder(csv_filename, reduce_output):
    """Test that the tagger decoder works with text output features when reduce_output is set to None."""
    input_features = [text_feature(encoder={"type": "parallel_cnn", "reduce_output": reduce_output})]
    output_features = [text_feature(output_feature=True, decoder={"type": "tagger"})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    with pytest.raises(ConfigValidationError) if reduce_output == "sum" else contextlib.nullcontext():
        run_experiment(input_features, output_features, dataset=rel_path)


@pytest.mark.parametrize("reduce_output", [("sum"), (None)], ids=["sum", "none"])
def test_experiment_sequence_output_feature_with_tagger_decoder(csv_filename, reduce_output):
    """Test that the tagger decoder works with sequence output features when reduce_output is set to None."""
    input_features = [text_feature(encoder={"type": "parallel_cnn", "reduce_output": reduce_output})]
    output_features = [sequence_feature(output_feature=True, decoder={"type": "tagger"})]

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    with pytest.raises(ConfigValidationError) if reduce_output == "sum" else contextlib.nullcontext():
        run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_category_input_feature_with_tagger_decoder(csv_filename):
    """Test that the tagger decoder doesn't work with category input features."""
    input_features = [category_feature()]
    output_features = [sequence_feature(output_feature=True, decoder={"type": "tagger"})]

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14, "reduce_output": None},
    }

    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    with pytest.raises(ConfigValidationError):
        run_experiment(config=config, dataset=rel_path)


def test_experiment_category_distribution_feature(csv_filename):
    vocab = ["a", "b", "c"]
    input_features = [vector_feature()]
    output_features = [
        category_distribution_feature(
            preprocessing={
                "vocab": vocab,
            }
        )
    ]
    # Generate test data
    rel_path = generate_data(input_features, output_features, csv_filename)

    input_df = pd.read_csv(rel_path)

    # set batch_size=auto to ensure we produce the correct shaped synthetic data
    config = {
        "input_features": input_features,
        "output_features": output_features,
        "combiner": {"type": "concat", "output_size": 14},
        TRAINER: {"epochs": 2, BATCH_SIZE: "auto"},
    }
    model, _, _, _, _ = run_experiment(input_features, output_features, dataset=rel_path, config=config)
    preds, _ = model.predict(input_df)

    # Check that predictions are category values drawn from the vocab, not distributions
    assert all(v in vocab for v in preds[f"{output_features[0][NAME]}_predictions"].values)


def test_experiment_ordinal_category(csv_filename):
    input_features = [category_feature(num_classes=5), number_feature()]
    output_features = [category_feature(output_feature=True, loss={"type": "corn"})]

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)


def test_experiment_feature_names_with_non_word_chars(tmpdir):
    config = yaml.safe_load(
        """
input_features:
    - name: Pclass (new)
      type: category
    - name: review.text
      type: category
    - name: other_feature
      type: category
      tied: review.text

output_features:
    - name: Survived (new)
      type: binary
    - name: Thrived
      type: binary
      dependencies:
        - Survived (new)

combiner:
    type: comparator
    entity_1:
        - Pclass (new)
        - other_feature
    entity_2:
        - review.text

"""
    )

    df = build_synthetic_dataset_df(120, config)
    model = LudwigModel(config, logging_level=logging.INFO)

    model.train(dataset=df, output_directory=tmpdir)


def test_text_output_feature_cols(tmpdir, csv_filename):
    """Test ensures that there are 4 output columns when model.predict() is called for text output features."""
    input_features = [text_feature(encoder={"type": "parallel_cnn"})]
    output_features = [text_feature(output_feature=True)]

    # Generate test data
    rel_path = generate_data(input_features, output_features, os.path.join(tmpdir, csv_filename))

    config = {
        "input_features": input_features,
        "output_features": output_features,
        "trainer": {"train_steps": 2, "batch_size": 5},
    }

    model = LudwigModel(config, logging_level=logging.INFO)
    model.train(dataset=rel_path, output_directory=tmpdir)
    predict_output = model.predict(dataset=rel_path)[0]

    assert len(predict_output.columns) == 4

    predict_df_headers = {col_name.split("_")[2] for col_name in list(predict_output.columns)}
    assert predict_df_headers == {"predictions", "probability", "probabilities", "response"}



from typing import Optional

from ludwig.api_annotations import DeveloperAPI
from ludwig.error import ConfigValidationError
from ludwig.schema import utils as schema_utils
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class RoPEScalingConfig(schema_utils.BaseMarshmallowConfig):
    """Dynamic RoPE-scaling (rotary position embeddings) to extend the context length of LLM like LLaMA, GPT-NeoX,
    or Falcon.

    This parameter is a dictionary containing the scaling configuration
    for the RoPE embeddings. Currently supports three scaling strategies: linear and dynamic. Their
    scaling factor must be an float greater than 1. The expected format is {'type': strategy name,
    'factor': scaling factor}
    """

    def __post_init__(self):
        # Both parameters must be set, or none.
        if not self.type:
            raise ConfigValidationError(
                f"`rope_scaling`'s `type` field must be one of ['linear', 'dynamic'], got {self.type}"
            )

        if not self.factor:
            raise ConfigValidationError(
                f"When using `rope_scaling`, `factor` must be specified and be > 1. Got {self.factor}."
            )

    type: Optional[str] = schema_utils.StringOptions(
        options=["linear", "dynamic"],
        default=None,
        allow_none=True,
        description="Currently supports two strategies: linear and dynamic scaling.",
    )

    factor: Optional[float] = schema_utils.FloatRange(
        default=None,
        allow_none=True,
        min=1.0,
        min_inclusive=False,
        description="The scaling factor for RoPE embeddings.",
    )


@DeveloperAPI
class RoPEScalingConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(RoPEScalingConfig, default_missing=True)

    def _jsonschema_type_mapping(self):
        return schema_utils.unload_jsonschema_from_marshmallow_class(RoPEScalingConfig, title="rope_scaling")


@DeveloperAPI
@ludwig_dataclass
class ModelParametersConfig(schema_utils.BaseMarshmallowConfig):
    rope_scaling: RoPEScalingConfig = RoPEScalingConfigField().get_default_field()

    def to_dict(self):
        config = {}
        if self.rope_scaling:
            config["rope_scaling"] = self.rope_scaling.to_dict()
        return config


@DeveloperAPI
class ModelParametersConfigField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(ModelParametersConfig, default_missing=True)

    def _jsonschema_type_mapping(self):
        return {
            "oneOf": [
                {"type": "null", "title": "disabled", "description": "Skip configurable model parameters."},
                {
                    **schema_utils.unload_jsonschema_from_marshmallow_class(ModelParametersConfig),
                    "title": "enabled",
                    "description": "Set model parameters options.",
                },
            ],
            "title": "Model Parameters",
            "description": "Configurable model parameters for LLMs.",
        }



import faiss
import numpy as np

from ludwig.vector_index.base import VectorIndex


class FaissIndex(VectorIndex):
    def __init__(self, index: faiss.Index):
        self.index = index

    def search(self, query: np.ndarray, k: int) -> np.ndarray:
        top_k = self.index.search(query.reshape(1, -1), k)
        return top_k[1].tolist()[0]

    def save(self, path: str):
        faiss.write_index(self.index, path)

    @classmethod
    def from_path(cls, path: str) -> "VectorIndex":
        index = faiss.read_index(path)
        return cls(index)

    @classmethod
    def from_embeddings(cls, embeddings: np.ndarray) -> "VectorIndex":
        index = faiss.IndexFlatL2(embeddings.shape[1])
        index.add(embeddings)
        return cls(index)



import logging
import os
import tempfile
from os import PathLike
from typing import Optional, Tuple, Type, Union

from transformers import AutoTokenizer, PreTrainedModel
from transformers.tokenization_utils import PreTrainedTokenizer

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.error_handling_utils import default_retry
from ludwig.utils.fs_utils import download, path_exists

logger = logging.getLogger(__name__)


@default_retry()
def load_pretrained_hf_model_from_hub(
    model_class: Type,
    pretrained_model_name_or_path: Optional[Union[str, PathLike]],
    **pretrained_kwargs,
) -> PreTrainedModel:
    """Download a HuggingFace model.

    Downloads a model from the HuggingFace zoo with retry on failure.
    Args:
        model_class: Class of the model to download.
        pretrained_model_name_or_path: Name of the model to download.
        pretrained_kwargs: Additional arguments to pass to the model constructor.
    Returns:
        The pretrained model object.
    """
    return model_class.from_pretrained(pretrained_model_name_or_path, **pretrained_kwargs)


@default_retry()
def load_pretrained_hf_tokenizer(
    pretrained_model_name_or_path: Optional[Union[str, PathLike]], **pretrained_kwargs
) -> PreTrainedTokenizer:
    """Download a HuggingFace tokenizer.

    Args:
        pretrained_model_name_or_path: Name of the tokenizer to download.
        pretrained_kwargs: Additional arguments to pass to the tokenizer constructor.
    Returns:
        The pretrained tokenizer object.
    """
    return AutoTokenizer.from_pretrained(pretrained_model_name_or_path, **pretrained_kwargs)


def _load_pretrained_hf_model_from_dir(
    model_class: Type,
    pretrained_model_name_or_path: Optional[Union[str, PathLike]],
    **pretrained_kwargs,
) -> PreTrainedModel:
    """Downloads a model to a local temporary directory, and Loads a pretrained HF model from a local directory."""
    with tempfile.TemporaryDirectory() as tmpdir:
        download(pretrained_model_name_or_path, tmpdir)
        return model_class.from_pretrained(tmpdir, **pretrained_kwargs)


@DeveloperAPI
def load_pretrained_hf_model_with_hub_fallback(
    model_class: Type,
    pretrained_model_name_or_path: Optional[Union[str, PathLike]],
    **pretrained_kwargs,
) -> Tuple[PreTrainedModel, bool]:
    """Returns the model and a boolean indicating whether the model was downloaded from the HuggingFace hub.

    If the `LUDWIG_PRETRAINED_MODELS_DIR` environment variable is set, we attempt to load the HF model from this
    directory, falling back to downloading from the HF hub if the model is not found, downloading fails, or if model
    initialization fails.

    `LUDWIG_PRETRAINED_MODELS_DIR` can be an s3 path. Weights are copied to a local temporary directory, and the model
    is loaded from there.

    The expected structure of the `LUDWIG_PRETRAINED_MODELS_DIR` directory is:
        {LUDWIG_PRETRAINED_MODELS_DIR}/{pretrained_model_name_or_path}/pytorch_model.bin
        {LUDWIG_PRETRAINED_MODELS_DIR}/{pretrained_model_name_or_path}/config.json

    For example, if `LUDWIG_PRETRAINED_MODELS_DIR` is set to `s3://my-bucket/pretrained-models`, and
    `pretrained_model_name_or_path` is set to `bert-base-uncased`, we expect to find the following files:
        s3://my-bucket/bert-base-uncased/
            - pytorch_model.bin
            - config.json

    If the `LUDWIG_PRETRAINED_MODELS_DIR` environment variable is not set, we download the model from the HF hub.
    """
    pretrained_models_dir = os.environ.get("LUDWIG_PRETRAINED_MODELS_DIR")
    if pretrained_models_dir:
        pretrained_model_path = os.path.join(pretrained_models_dir, pretrained_model_name_or_path)
        if path_exists(pretrained_model_path):
            try:
                logger.info(
                    f"Found existing pretrained model artifact {pretrained_model_name_or_path} in directory "
                    f"{pretrained_models_dir}. Downloading."
                )
                return (
                    _load_pretrained_hf_model_from_dir(model_class, pretrained_model_path, **pretrained_kwargs),
                    False,
                )
            except Exception as e:
                logger.warning(
                    f"Failed to download pretrained model from {pretrained_models_dir} with error {e}. "
                    "Falling back to HuggingFace model hub."
                )

    # Fallback to HF hub.
    return load_pretrained_hf_model_from_hub(model_class, pretrained_model_name_or_path, **pretrained_kwargs), True



from typing import Any, Dict, Type, Union

from ludwig.distributed.base import DistributedStrategy, LocalStrategy


def load_ddp():
    from ludwig.distributed.ddp import DDPStrategy

    return DDPStrategy


def load_fsdp():
    from ludwig.distributed.fsdp import FSDPStrategy

    return FSDPStrategy


def load_deepspeed():
    from ludwig.distributed.deepspeed import DeepSpeedStrategy

    return DeepSpeedStrategy


def load_horovod():
    from ludwig.distributed.horovod import HorovodStrategy

    return HorovodStrategy


def load_local():
    return LocalStrategy


STRATEGIES = {
    "ddp": load_ddp,
    "fsdp": load_fsdp,
    "deepspeed": load_deepspeed,
    "horovod": load_horovod,
    "local": load_local,
}


_current_strategy: DistributedStrategy = None


def init_dist_strategy(strategy: Union[str, Dict[str, Any]], **kwargs) -> DistributedStrategy:
    global _current_strategy
    if isinstance(strategy, dict):
        dtype = strategy.pop("type", None)
        obj = get_dist_strategy(dtype)(**strategy)
    else:
        obj = get_dist_strategy(strategy)(**kwargs)
    _current_strategy = obj
    return obj


def get_current_dist_strategy() -> DistributedStrategy:
    if _current_strategy is None:
        raise RuntimeError("Distributed strategy not initialized")
    return _current_strategy


def get_dist_strategy(strategy: Union[str, Dict[str, Any]]) -> Type[DistributedStrategy]:
    name = strategy
    if isinstance(strategy, dict):
        name = strategy["type"]
    return STRATEGIES[name]()


def get_default_strategy_name() -> str:
    return "ddp"



from typing import Any, Dict

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import CATEGORY, MODEL_LLM, TEXT
from ludwig.schema import utils as schema_utils
from ludwig.schema.decoders.base import BaseDecoderConfig
from ludwig.schema.decoders.utils import register_decoder_config
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class BaseExtractorDecoderConfig(BaseMarshmallowConfig):
    tokenizer: str = "hf_tokenizer"

    input_size: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Size of the input to the decoder.",
    )

    pretrained_model_name_or_path: str = schema_utils.String(
        default="",
        allow_none=True,
        description="Path to the pretrained model or model identifier from huggingface.co/models.",
    )

    vocab_file: str = schema_utils.String(
        default="",
        allow_none=True,
        description="Path to the vocabulary file.",
    )

    max_new_tokens: int = schema_utils.Integer(
        default=None,
        allow_none=True,
        description="Maximum number of new tokens that will be generated.",
    )


@DeveloperAPI
@register_decoder_config("text_extractor", [TEXT], model_types=[MODEL_LLM])
@ludwig_dataclass
class TextExtractorDecoderConfig(BaseExtractorDecoderConfig, BaseDecoderConfig):
    @classmethod
    def module_name(cls):
        return "TextExtractorDecoder"

    type: str = schema_utils.ProtectedString("text_extractor")


@DeveloperAPI
@register_decoder_config("category_extractor", [CATEGORY], model_types=[MODEL_LLM])
@ludwig_dataclass
class CategoryExtractorDecoderConfig(BaseExtractorDecoderConfig, BaseDecoderConfig):
    @classmethod
    def module_name(cls):
        return "CategoryExtractorDecoder"

    type: str = schema_utils.ProtectedString("category_extractor")

    # Match is a dict of label class
    match: Dict[str, Dict[str, Any]] = schema_utils.Dict(
        default=None,
        allow_none=False,
        description="A dictionary of label classes and their corresponding "
        "match patterns definitions that will be used to parse the output "
        "of the LLM.",
    )

    str2idx: Dict[str, int] = schema_utils.Dict(
        default=None,
        allow_none=True,
        description="A dictionary of label classes and their corresponding "
        "indices that will be used to parse the output of the LLM.",
    )

    fallback_label: str = schema_utils.String(
        default="",
        allow_none=True,
        description="The label to use if the parser fails to parse the input.",
    )



from collections import OrderedDict
from typing import Dict, List, Tuple

import numpy as np
import torch

NUMPY_TO_TORCH_DTYPE = {
    bool: torch.bool,
    np.bool_: torch.bool,
    np.uint8: torch.uint8,
    np.int8: torch.int8,
    np.int16: torch.int16,
    np.int32: torch.int32,
    np.int64: torch.int64,
    np.float16: torch.float16,
    np.float32: torch.float32,
    np.float64: torch.float64,
    np.complex64: torch.complex64,
    np.complex128: torch.complex128,
}


def extract_tensors(model: torch.nn.Module) -> Tuple[torch.nn.Module, List[Dict]]:
    """Remove the tensors from a PyTorch model, convert them to NumPy arrays, and return the stripped model and
    tensors.

    Reference implementation: https://medium.com/ibm-data-ai/how-to-load-pytorch-models-340-times-faster-with-
    ray-8be751a6944c  # noqa
    """

    tensors = []
    for _, module in model.named_modules():
        # Store the tensors as numpy arrays in Python dictionaries
        # Delete the same tensors since we no longer need them and we want to reduce memory pressure.
        # This ensures that throughout this process, we keep memory nearly linear w.r.t model parameters.
        params = OrderedDict()
        buffers = OrderedDict()
        for name, param in module.named_parameters(recurse=False):
            params[name] = torch.clone(param).detach().numpy()
            del param
        for name, buf in module.named_buffers(recurse=False):
            buffers[name] = torch.clone(buf).detach().numpy()
            del buf
        tensors.append({"params": params, "buffers": buffers})

    # Strip all tensors and buffers out of the original model.
    for _, module in model.named_modules():
        for name in [name for name, _ in module.named_parameters(recurse=False)] + [
            name for name, _ in module.named_buffers(recurse=False)
        ]:
            setattr(module, name, None)

    return model, tensors


def replace_tensors(m: torch.nn.Module, tensors: List[Dict], device: torch.device):
    """Restore the tensors that extract_tensors() stripped out of a PyTorch model. This operation is performed in
    place.

    Reference implementation: https://medium.com/ibm-data-ai/how-to-load-pytorch-models-340-times-faster-with-
    ray-8be751a6944c  # noqa
    """
    modules = [module for _, module in m.named_modules()]
    for module, tensor_dict in zip(modules, tensors):
        # There are separate APIs to set parameters and buffers.
        for name, array in tensor_dict["params"].items():
            module.register_parameter(
                name,
                torch.nn.Parameter(torch.as_tensor(array, device=device, dtype=NUMPY_TO_TORCH_DTYPE.get(array.dtype))),
            )

        for name, array in tensor_dict["buffers"].items():
            module.register_buffer(
                name,
                torch.as_tensor(array, device=device, dtype=NUMPY_TO_TORCH_DTYPE.get(array.dtype)),
            )


def find_embedding_layer_with_path(module, module_names=[]):
    """Recursively search through a module to find an embedding layer and its module path.

    Returns a tuple containing the embedding layer and its module path.
    """
    for name, child_module in module.named_children():
        if isinstance(child_module, torch.nn.Embedding):
            # If an embedding layer is found, return it along with the module path
            return child_module, ".".join(module_names + [name])
        else:
            # Recursively search in the child module and update the module_names list
            found, path = find_embedding_layer_with_path(child_module, module_names + [name])
            if found is not None:
                return found, path
    return None, None



import torch
from transformers import AutoModelForCausalLM

from ludwig.utils.model_utils import extract_tensors, find_embedding_layer_with_path, replace_tensors

# Define a sample model for testing


class SampleModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.conv = torch.nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1)
        self.relu = torch.nn.ReLU()


def test_extract_tensors():
    # Create a sample model
    model = SampleModel()

    # Call extract_tensors function
    stripped_model, tensors = extract_tensors(model)

    # Assert that the model and tensors are returned
    assert isinstance(stripped_model, torch.nn.Module)
    assert isinstance(tensors, list)

    # Assert that the tensors contain the expected keys
    for tensor_dict in tensors:
        assert "params" in tensor_dict
        assert "buffers" in tensor_dict

    # Assert that all model parameters are set to None
    for module in stripped_model.modules():
        for name, param in module.named_parameters(recurse=False):
            assert param is None

        for name, buf in module.named_buffers(recurse=False):
            assert buf is None


def test_replace_tensors():
    # Create a sample model
    model = SampleModel()

    # Call extract_tensors function to get the tensors
    _, tensors = extract_tensors(model)

    # Create a new device for testing
    device = torch.device("cpu")

    # Call replace_tensors function
    replace_tensors(model, tensors, device)

    # Assert that the tensors are restored
    for module, tensor_dict in zip(model.modules(), tensors):
        for name, array in tensor_dict["params"].items():
            assert name in module._parameters
            assert torch.allclose(module._parameters[name], torch.as_tensor(array, device=device))

        for name, array in tensor_dict["buffers"].items():
            assert name in module._buffers
            assert torch.allclose(module._buffers[name], torch.as_tensor(array, device=device))


# Define a sample module structure for testing
class SampleModule(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.embedding = torch.nn.Embedding(10, 20)
        self.rnn = torch.nn.LSTM(20, 30)


def test_find_embedding_layer_with_path_simple():
    # Test case 1: Test the function with a simple module structure
    module = SampleModule()
    embedding_layer, path = find_embedding_layer_with_path(module)
    assert embedding_layer is not None
    assert isinstance(embedding_layer, torch.nn.Embedding)
    assert path == "embedding"


def test_find_embedding_layer_with_path_complex():
    # Test case 2: Test the function with a more complex module structure including AutoModelForCausalLM
    model = AutoModelForCausalLM.from_pretrained("HuggingFaceM4/tiny-random-LlamaForCausalLM")

    embedding_layer, path = find_embedding_layer_with_path(model)
    assert embedding_layer is not None
    assert isinstance(embedding_layer, torch.nn.Embedding)
    assert path == "model.embed_tokens"


def test_no_embedding_layer():
    # Test case 3: Embedding layer is not present
    no_embedding_model = torch.nn.Sequential(torch.nn.Linear(10, 10), torch.nn.Linear(10, 10))
    embedding_layer, path = find_embedding_layer_with_path(no_embedding_model)
    assert embedding_layer is None
    assert path is None



#! /usr/bin/env python
# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from abc import ABC, abstractmethod

from torch import nn

from ludwig.api_annotations import DeveloperAPI
from ludwig.utils.torch_utils import LudwigModule


@DeveloperAPI
class Encoder(LudwigModule, ABC):
    @abstractmethod
    def forward(self, inputs, training=None, mask=None):
        raise NotImplementedError

    def get_embedding_layer(self) -> nn.Module:
        """Returns layer that embeds inputs, used for computing explanations.

        Captum adds an evaluation hook to this module returned by this function. The hook copies the module's return
        with .clone(). The module returned by this function must return a tensor, not a dictionary of tensors.
        """
        return next(self.children())

    @property
    def name(self) -> str:
        return self.__class__.__name__



from copy import deepcopy
from datetime import date, datetime
from typing import Any, List

import pytest
import torch
from dateutil.parser import parse

from ludwig.constants import ENCODER_OUTPUT, FILL_WITH_CONST, MISSING_VALUE_STRATEGY
from ludwig.features import date_feature
from ludwig.features.date_feature import DateInputFeature
from ludwig.schema.features.date_feature import DateInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from ludwig.types import FeatureConfigDict
from ludwig.utils.date_utils import create_vector_from_datetime_obj
from ludwig.utils.misc_utils import merge_dict
from ludwig.utils.torch_utils import get_torch_device

BATCH_SIZE = 2
DATE_W_SIZE = 9
DEVICE = get_torch_device()


@pytest.fixture(scope="module")
def date_config():
    return {"name": "date_column_name", "type": "date"}


def test_date_input_feature(date_config: FeatureConfigDict):
    # setup image input feature definition
    feature_def = deepcopy(date_config)

    # pickup any other missing parameters
    defaults = DateInputFeatureConfig(name="foo").to_dict()
    set_def = merge_dict(defaults, feature_def)

    # ensure no exceptions raised during build
    feature_config, _ = load_config_with_kwargs(DateInputFeatureConfig, set_def)
    input_feature_obj = DateInputFeature(feature_config).to(DEVICE)

    # check one forward pass through input feature
    input_tensor = input_feature_obj.create_sample_input(batch_size=BATCH_SIZE)
    assert input_tensor.shape == torch.Size((BATCH_SIZE, DATE_W_SIZE))
    assert input_tensor.dtype == torch.int32

    encoder_output = input_feature_obj(input_tensor)
    assert encoder_output[ENCODER_OUTPUT].shape == (BATCH_SIZE, *input_feature_obj.output_shape)


@pytest.mark.parametrize(
    "date_str,datetime_format,expected_list",
    [
        ("2012-02-26T13:51:50.417-07:00", None, [2012, 2, 26, 6, 57, 13, 51, 50, 49910]),
        ("2022-06-25 09:30:59", None, [2022, 6, 25, 5, 176, 9, 30, 59, 34259]),
        ("2022-06-25", None, [2022, 6, 25, 5, 176, 0, 0, 0, 0]),
    ],
)
def test_date_to_list(date_str, datetime_format, expected_list):
    preprocessing_parameters = None
    assert (
        date_feature.DateInputFeature.date_to_list(date_str, datetime_format, preprocessing_parameters) == expected_list
    )


@pytest.fixture(scope="module")
def reference_date_list() -> List[int]:
    return create_vector_from_datetime_obj(datetime.utcfromtimestamp(1691600953.443032))


@pytest.fixture(scope="module")
def fill_value() -> str:
    return "1970-01-01 00:00:00"


@pytest.fixture(scope="module")
def fill_value_list(fill_value: str) -> List[int]:
    return create_vector_from_datetime_obj(parse(fill_value))


@pytest.mark.parametrize(
    "timestamp,datetime_format,expected_list",
    [
        pytest.param(1691600953.443032, None, "reference_date_list", id="float-s"),
        pytest.param(1691600953443.032, None, "reference_date_list", id="float-ms"),
        pytest.param(1691600953, None, "reference_date_list", id="int-s"),
        pytest.param(1691600953443, None, "reference_date_list", id="int-ms"),
        pytest.param(1691600953.443032, "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="float-s-fmt"),
        pytest.param(1691600953443.032, "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="float-ms-fmt"),
        pytest.param(1691600953, "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="int-s-fmt"),
        pytest.param(1691600953443, "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="int-ms-fmt"),
        pytest.param("1691600953.443032", None, "reference_date_list", id="string[float]-s"),
        pytest.param("1691600953443.0032", None, "reference_date_list", id="string[float]-ms"),
        pytest.param("1691600953", None, "reference_date_list", id="string[int]-s"),
        pytest.param("1691600953443", None, "reference_date_list", id="string[int]-ms"),
        pytest.param("1691600953.443032", "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="string[float]-s-fmt"),
        pytest.param("1691600953443.0032", "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="string[float]-ms-fmt"),
        pytest.param("1691600953", "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="string[int]-s-fmt"),
        pytest.param("1691600953443", "%d/%m/%y %H:%M:%S.%f", "reference_date_list", id="string[int]-ms-fmt"),
        pytest.param("foo", None, "fill_value_list", id="string error"),
        pytest.param([1691600953.443032], None, "fill_value_list", id="list error"),
        pytest.param(None, None, "fill_value_list", id="NoneType error"),
    ],
)
def test_date_to_list_numeric(timestamp: Any, datetime_format: str, expected_list: List[int], fill_value: str, request):
    """Test that numeric datetime formats are converted correctly.

    Currently, we support int, float, and string representations of POSIX timestamps in seconds and milliseconds. Valid
    timestamps should be converted to datetime lists by `luwdig.utils.date_utils.create_vector_from_datetime_object`.
    If a string format is provided, it should be ignored.

    Args:
        timestamp: Input to be converted to a date vector
        datetime_format: Optional format string, should be ignored under the hood with these timestamps.
        expected_list: The expected output of `DateFeatureMixin.date_to_list`
        fill_value: Date to be used as fallback
        request: pytest request fixture
    """
    expected_result = request.getfixturevalue(expected_list)

    # The default fill value is `datetime.now`, for testing we override this to be a constant.
    preprocessing_parameters = {MISSING_VALUE_STRATEGY: FILL_WITH_CONST, "fill_value": fill_value}

    # No exception should ever be raised from `date_to_list` due to a parsing error. The expected behavior is to fall
    # back to the fill value.
    dt = date_feature.DateInputFeature.date_to_list(timestamp, datetime_format, preprocessing_parameters)
    assert dt == expected_result


def test_date_to_list__DatetimeObjectFromParsedJSON():
    preprocessing_parameters = None
    datetime_obj = datetime.fromisoformat("2022-06-25")
    assert date_feature.DateInputFeature.date_to_list(datetime_obj, None, preprocessing_parameters) == [
        2022,
        6,
        25,
        5,
        176,
        0,
        0,
        0,
        0,
    ]


def test_date_to_list__UsesFillValueOnInvalidDate():
    preprocessing_parameters = {"fill_value": "2013-02-26"}
    invalid_date_str = "2012abc-02"
    datetime_format = None
    assert date_feature.DateInputFeature.date_to_list(invalid_date_str, datetime_format, preprocessing_parameters) == [
        2013,
        2,
        26,
        1,
        57,
        0,
        0,
        0,
        0,
    ]


@pytest.fixture(scope="module")
def date_obj():
    return date.fromisoformat("2022-06-25")


@pytest.fixture(scope="module")
def date_obj_vec():
    return create_vector_from_datetime_obj(datetime.fromisoformat("2022-06-25"))


def test_date_object_to_list(date_obj, date_obj_vec, fill_value):
    """Test support for datetime.date object conversion.

    Args:
        date_obj: Date object to convert into a vector
        date_obj_vector: Expected vector version of `date_obj`
    """
    computed_date_vec = date_feature.DateInputFeature.date_to_list(
        date_obj, None, preprocessing_parameters={MISSING_VALUE_STRATEGY: FILL_WITH_CONST, "fill_value": fill_value}
    )
    assert computed_date_vec == date_obj_vec
